import os
import numpy as np
import subprocess
import mmap
import warnings
import h5py
import re

from mcstasscript.helper.formatting import bcolors
from mcstasscript.data.data import McStasMetaData
from mcstasscript.data.data import McStasDataBinned
from mcstasscript.data.data import McStasDataEvent


class ManagedMcrun:
    """
    A class for performing a mcstas simulation and organizing the data
    into python objects

    ManagedMcrun is usually called by the instrument class of
    McStasScript but can be used independently.  It runs the mcrun
    command using the system command, and if this is not in the path,
    the absolute path can be given in a keyword argument executable_path.

    Attributes
    ----------
    name_of_instrumentfile : str
        Name of instrument file to be executed

    data_folder_name : str
        Name of datafolder mcrun writes to disk

    ncount : int
        Number of rays to simulate

    mpi : int
        Number of mpi threads to run

    parameters : dict
        Dictionary of parameter names and values for this simulation

    custom_flags : string
        Custom flags that are passed to the mcrun command

    executable_path : string
        Path to the mcrun command (can be empty if already in path)

    Methods
    -------
    run_simulation()
        Runs simulation, returns list of McStasData instances

    """

    def __init__(self, instr_name, **kwargs):
        """
        Performs call to McStas with given options

        Uses subprocess to call mcrun / mxrun to perform simulation of given
        instrument file.

        Parameters
        ----------
        instr_name : str
            Name of instrument file to be simulated

        kwargs : keyword arguments
            output_path : str, required
                Sets data_folder_name
            ncount : int, default 1E6
                Sets ncount
            mpi : int, default None
                Sets thread count, None to disable mpi
            gravity : bool, default False
                Enables gravity if True
            parameters : dict
                Sets parameters
            custom_flags : str, default ""
                Sets custom_flags passed to mcrun
            executable_path : str
                Path to mcrun command, "" if already in path
            increment_folder_name : bool, default True
                If True, automatically appends output_path to make it unique
            force_compile : bool, default True
                If True, forces compile. If False no new instrument is written
            run_folder : str
                Path to folder in which to run McStas
            openacc : bool, default False
                If True, adds the --openacc flag to mcrun call
            NeXus : bool, default False
                If True, adds the --format=NeXus to mcrun call

        """

        self.name_of_instrumentfile = instr_name

        self.data_folder_name = ""
        self.ncount = int(1E6)
        self.mpi = None
        self.gravity = False
        self.openacc = False
        self.NeXus = False
        self.parameters = {}
        self.custom_flags = ""
        self.executable_path = ""
        self.executable = ""
        self.increment_folder_name = True
        self.compile = True
        self.run_path = "."
        self.seed = None
        self.suppress_output = False

        # executable_path always in kwargs
        if "executable_path" in kwargs:
            self.executable_path = kwargs["executable_path"]

        if "executable" in kwargs:
            self.executable = kwargs["executable"]

        if "output_path" in kwargs:
            self.data_folder_name = kwargs["output_path"]
        else:
            raise NameError(
                "ManagedMcrun needs output_path to load data, add "
                + "with keyword argument.")

        if "ncount" in kwargs:
            self.ncount = int(kwargs["ncount"])

            if self.ncount < 1:
                raise ValueError("ncount should be a positive integer, was "
                                 + str(self.ncount))

        if "mpi" in kwargs:
            self.mpi = kwargs["mpi"]
            try:
                self.mpi = int(self.mpi)
            except (TypeError, ValueError) as e:
                if self.mpi is not None:
                    raise RuntimeError("MPI should be an integer, was "
                                       + str(self.mpi))

            if self.mpi is not None:
                if self.mpi < 1:
                    raise ValueError("MPI should be an integer larger than"
                                     + " 0, was " + str(self.mpi))

        if "gravity" in kwargs:
            self.gravity = kwargs["gravity"]

        if "openacc" in kwargs:
            self.openacc = kwargs["openacc"]

        if "NeXus" in kwargs:
            self.NeXus = kwargs["NeXus"]

        if "seed" in kwargs:
            self.seed = kwargs["seed"]

        if "parameters" in kwargs:
            self.parameters = kwargs["parameters"]

            if not isinstance(self.parameters, dict):
                raise RuntimeError("Parameters should be given as dict.")

        if "custom_flags" in kwargs:
            self.custom_flags = kwargs["custom_flags"]

            if not isinstance(self.custom_flags, str):
                raise RuntimeError("ManagedMcrun detected given customf_flags"
                                   + " was not a string.")

        if "increment_folder_name" in kwargs:
            self.increment_folder_name = kwargs["increment_folder_name"]

        if "force_compile" in kwargs:
            self.compile = kwargs["force_compile"]

        if "run_path" in kwargs:
            self.run_path = kwargs["run_path"]

        if "suppress_output" in kwargs:
            self.suppress_output = bool(kwargs["suppress_output"])


        # get relevant paths and check their validity
        current_directory = os.getcwd()

        if not os.path.isabs(self.data_folder_name):
            self.data_folder_name = os.path.join(current_directory,
                                                 self.data_folder_name)
        else:
            split_data_path = os.path.split(self.data_folder_name)
            if not os.path.isdir(split_data_path[0]):
                raise RuntimeError("Parent folder for datafolder invalid: "
                                   + str(split_data_path[0]))

        if not os.path.isabs(self.run_path):
            self.run_path = os.path.join(current_directory, self.run_path)
        else:
            split_run_path = os.path.split(self.run_path)
            if not os.path.isdir(split_run_path[0]):
                raise RuntimeError("Parent folder for run_path invalid: "
                                   + str(split_run_path[0]))

        if not os.path.isdir(self.run_path):
            raise RuntimeError("ManagedMcrun found run_path to "
                               + "be invalid: " + str(self.run_path))

        if not os.path.isdir(self.executable_path):
            raise RuntimeError("ManagedMcrun found executable_path to "
                               + "be invalid: " + str(self.executable_path))

    def run_simulation(self):
        """
        Runs McStas simulation described by initializing the object
        """

        # construct command to run
        option_string = ""
        if self.compile:
            option_string += "-c "

        if self.gravity:
            option_string += "-g "

        if self.NeXus:
            option_string += "--format=NeXus "

        if self.openacc:
            option_string += "--openacc "

        if self.mpi is not None:
            mpi_string = "--mpi=" + str(self.mpi) + " "  # Set mpi
        else:
            mpi_string = ""

        if self.seed is not None:
            seed_string = "--seed=" + str(self.seed) + " " # Set seed
        else:
            seed_string = ""

        option_string = (option_string
                         + "-n " + str(self.ncount) + " " # Set ncount
                         + mpi_string
                         + seed_string)

        if os.path.exists(self.data_folder_name):
            if self.increment_folder_name:
                counter = 0
                new_name = self.data_folder_name + "_" + str(counter)
                while os.path.isdir(new_name):
                    counter = counter + 1
                    new_name = self.data_folder_name + "_" + str(counter)

                self.data_folder_name = new_name
            else:
                raise NameError("output_path already exists and "
                                + "increment_folder_name was set to False.")

        if len(self.data_folder_name) > 0:
            option_string = (option_string
                             + "-d "
                             + self.data_folder_name)

        # add parameters to command
        parameter_string = ""
        for key, val in self.parameters.items():
            parameter_string = (parameter_string + " "
                                + str(key)  # parameter name
                                + "="
                                + str(val))  # parameter value

        mcrun_full_path = os.path.join(self.executable_path, self.executable)
        if len(self.executable_path) > 1:
            if not (self.executable_path[-1] == "\\"
                    or self.executable_path[-1] == "/"):
                mcrun_full_path = os.path.join(self.executable_path,
                                               self.executable)
        mcrun_full_path = '"' + mcrun_full_path + '"' # Path in quotes to allow spaces

        # Run the mcrun command on the system
        full_command = (mcrun_full_path + " "
                        + option_string + " "
                        + self.custom_flags + " "
                        + self.name_of_instrumentfile
                        + parameter_string)

        process = subprocess.run(full_command, shell=True,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.STDOUT,
                                 universal_newlines=True,
                                 cwd=self.run_path)

        if self.suppress_output is False:
            print_sim_output(process.stdout)

        if not os.path.isdir(self.data_folder_name):
            warnings.warn("Simulation did not create data folder, most likely failed.")

    def load_results(self, *args):
        """
        Method for loading data from a mcstas simulation

        Loads data on all monitors in a McStas data folder, and returns these
        as a list of McStasData objects.

        Parameters
        ----------

        optional first argument : str
            path to folder from which data should be loaded

        """

        if len(args) == 0:
            data_folder_name = self.data_folder_name
        elif len(args) == 1:
            data_folder_name = args[0]
        else:
            raise RuntimeError("load_results can be called "
                               + "with 0 or 1 arguments")

        if os.path.isdir(data_folder_name):
            return load_results(data_folder_name)
        else:
            warnings.warn("No data available to load.")
            return None


def load_results(data_folder_name):
    """
    Function for loading data from a mcstas simulation

    Loads data on all monitors in a McStas data folder, and returns these
    as a list of McStasData objects.

    Parameters
    ----------

    data_folder_name : str
        path to folder from which data should be loaded

    """

    if not os.path.isdir(data_folder_name):
        raise NameError("Given data directory does not exist.")

    # Find all data files in generated folder
    files_in_folder = os.listdir(data_folder_name)

    # Raise an error if mccode.sim is not available
    if "mccode.sim" in files_in_folder:
        NeXus = False
    elif "mccode.h5" in files_in_folder:
        NeXus = True
    else:
        raise NameError("No mccode.sim or mccode.h5 in data folder.")

    if NeXus:
        # Open mccode to read metadata for all datasets written to disk
        with h5py.File(os.path.join(data_folder_name, "mccode.h5"), "r", swmr=True) as f:

            # Pass file object to all functions to avoid multiple open / close
            metadata_list = load_metadata_nexus(f)

            results = []
            for metadata in metadata_list:
                result = load_monitor_nexus(metadata, f)
                result.set_data_location(data_folder_name)
                results.append(result)

    else:
        # Older workflow, still handles both text and NeXus
        metadata_list = load_metadata(data_folder_name)

        results = []
        for metadata in metadata_list:
            result = load_monitor(metadata, data_folder_name)
            result.set_data_location(data_folder_name)
            results.append(result)

    return results

def load_metadata(data_folder_name):
    """
    Function that loads metadata from a mcstas simulation

    Returns list of metadata objects corresponding to each monitor, all
    information is taken from mccode.sim file.

    Parameters
    ----------

    first argument : str
        path to folder from which metadata should be loaded
    """

    if not os.path.isdir(data_folder_name):
        raise NameError("Given data directory does not exist.")

    # Find all data files in generated folder
    files_in_folder = os.listdir(data_folder_name)

    # Raise an error if mccode.sim is not available
    if "mccode.sim" in files_in_folder:
        return load_metadata_text(data_folder_name)
    elif "mccode.h5" in files_in_folder:
        with h5py.File(os.path.join(data_folder_name, "mccode.h5"), "r", swmr=True) as f:
            return load_metadata_nexus(f)
    else:
        raise NameError("No mccode.sim or mccode.h5 in data folder.")


def load_metadata_text(data_folder_name):
    instrument_parameters = {}

    # Open mccode to read metadata for all datasets written to disk
    with open(os.path.join(data_folder_name, "mccode.sim"), "r") as f:

        # Loop that reads mccode.sim sections
        metadata_list = []
        current_object = None
        in_data = False
        in_sim = False
        for lines in f:
            # Could read other details about run

            if lines == "end data\n":
                # No more data for this metadata object
                # Add parameter information
                current_object.add_info("Parameters", instrument_parameters)
                # Extract the information
                current_object.extract_info()
                # Add to metadata list
                if current_object.filename != "":
                    metadata_list.append(current_object)
                # Stop reading data
                in_data = False

            if in_sim:
                if "Param" in lines:
                    parm_lst = lines.split(':')[1].split('=')
                    try:
                        value = float(parm_lst[1].strip())
                    except ValueError:
                        value = parm_lst[1].strip()

                    instrument_parameters[parm_lst[0].strip()] = value

            if in_data:
                # This line contains info to be added to metadata
                colon_index = lines.index(":")
                key = lines[2:colon_index]
                value = lines[colon_index + 2:].strip()
                current_object.add_info(key, value)

            if lines == "begin data\n":
                # Found data section, create new metadata object
                current_object = McStasMetaData()
                # Start recording data to metadata object
                in_data = True

            if 'begin simulation:' in lines:
                in_sim = True
            if 'end simulation:' in lines:
                in_sim = False

        # Close mccode.sim
        f.close()

        """
        # Create a list for McStasData instances to return
        results = []


        # Load datasets described in metadata list individually
        for metadata in metadata_list:

            # Load data with numpy
            data = np.loadtxt(data_folder_name
                              + "/"
                              + metadata.filename.rstrip())

            # Split data into intensity, error and ncount
            if type(metadata.dimension) == int and metadata.dimension == 0:
                Intensity = data.T

            if type(metadata.dimension) == int and metadata.dimension != 0:

                xaxis = data.T[0, :]
                Intensity = data.T[1, :]
                Error = data.T[2, :]
                Ncount = data.T[3, :]

            elif len(metadata.dimension) == 2:
                xaxis = []  # Assume evenly binned in 2d
                data_lines = metadata.dimension[1]
                Intensity = data.T[:, 0:data_lines - 1]
                Error = data.T[:, data_lines:2*data_lines - 1]
                Ncount = data.T[:, 2*data_lines:3*data_lines - 1]
        """

    return metadata_list


def load_metadata_nexus(file_object):
    instrument_parameters = {}

    f = file_object

    if "entry1" not in list(f.keys()):
        raise ValueError("h5 file not formatted as expected.")

    if "data" not in list(f["entry1"].keys()):
        raise ValueError("h5 file not formatted as expected.")

    if "simulation" not in list(f["entry1"].keys()):
        raise ValueError("h5 file not formatted as expected.")

    if "Param" not in list(f["entry1"]["simulation"].keys()):
        raise ValueError("h5 file not formatted as expected.")

    # Common information

    # Instrument parameters
    instrument_parameters = {}
    loaded_par_dict = f["entry1"]["simulation"]["Param"].attrs
    for par_name in loaded_par_dict:
        if par_name == "NX_class":
            continue

        try:
            value = float(loaded_par_dict[par_name])
        except:
            value = str(loaded_par_dict[par_name])

        instrument_parameters[par_name] = value

    metadata_list = []

    # For each entry in data, make a metadata object
    for key in f["entry1"]["data"].keys():

        # Make the metadata object and add instrument parameters
        metadata = McStasMetaData()
        metadata.add_info("Parameters", instrument_parameters)

        # Add NeXus field name
        metadata.add_info("NeXus_field", key)

        # Add all the read info from attribute section
        info = dict(f["entry1"]["data"][key].attrs)
        info = decode_dict(info)
        for name, value in info.items():
            if isinstance(value, bytes):
                value = value.decode('utf-8')

            metadata.add_info(name, value)

        metadata_list.append(metadata)

        # Now all info is added, extract info loads it into nice attributes
        metadata.extract_info()

    return metadata_list


def decode_dict(dictionary):
    for key, value in dictionary.items():
        if isinstance(value, bytes):
            try:
                dictionary[key] = value.decode('utf-8')
            except:
                # Investigate cases where this fail when reading from nexus
                dictionary[key] = value.decode('utf-8', errors='replace')  # Replaces invalid bytes with '?'

    return dictionary


def load_monitor(metadata, data_folder_name):
    """
    Switches to appropriate loader function
    """

    if "NeXus_field" in metadata.info:
        with h5py.File(os.path.join(data_folder_name, "mccode.h5"), "r", swmr=True) as f:
            return load_monitor_nexus(metadata, f)
    else:
        return load_monitor_text(metadata, data_folder_name)


def load_monitor_nexus(metadata, file_object):
    """
    Function that loads data given metadata and name of data folder
    This version is for a nexus file

    Loads data for single monitor and returns a McStasData object

    Parameters
    ----------

    metadata : McStasMetaData object
        McStasMetaData object corresponding to the monitor to be loaded

    file_object : h5py file object in read mode
    """

    f = file_object

    if "entry1" not in list(f.keys()):
        raise ValueError("h5 file not formatted as expected.")

    if "data" not in list(f["entry1"].keys()):
        raise ValueError("h5 file not formatted as expected.")

    if "simulation" not in list(f["entry1"].keys()):
        raise ValueError("h5 file not formatted as expected.")

    if "Param" not in list(f["entry1"]["simulation"].keys()):
        raise ValueError("h5 file not formatted as expected.")

    NeXus_field = metadata.info["NeXus_field"]

    available_fields = f["entry1"]["data"][NeXus_field].keys()
    if not metadata.dimension == 0 and "events" not in available_fields:
        if "data" not in available_fields:
            raise ValueError("NeXus reading: data not found! \n"
                             + "Monitor metadata:\n" + str(metadata))

        if "errors" not in available_fields:
            raise ValueError("NeXus reading: errors not found! \n"
                             + "Monitor metadata:\n" + str(metadata))

        if "ncount" not in available_fields:
            raise ValueError("NeXus reading: ncount not found! \n"
                             + "Monitor metadata:\n" + str(metadata))

    # Need to check if it is binned data or event data
    if "events" in available_fields:
        Events = np.array(f["entry1"]["data"][NeXus_field]["events"])
        return McStasDataEvent(metadata, Events)

    # Split data into intensity, error and ncount
    if type(metadata.dimension) == int and metadata.dimension == 0:

        if "data" in f["entry1"]["data"][NeXus_field].keys():
            raise ValueError("Found array data in 0D dataset?")

        values = None
        if "values" in f["entry1"]["data"][NeXus_field].keys():
            values = np.array(f["entry1"]["data"][NeXus_field]["values"])

        if metadata.total_I is None:
            if values is not None:
                Intensity = np.array(values[0])
            else:
                raise ValueError("No info on this monitor can be found "
                                 + "in reading of NeXus file "
                                 + str(metadata))
        else:
            Intensity = np.array(metadata.total_I)

        if metadata.total_E is None:
            if values is not None:
                Error = np.array(values[2])
            else:
                Error = np.zeros(1)
        else:
            Error = np.array(metadata.total_E)

        if metadata.total_N is None:
            if values is not None:
                Ncount = np.array(values[3])
            else:
                Ncount = np.zeros(1)
        else:
            Ncount = np.array(metadata.total_N)

        return McStasDataBinned(metadata, Intensity, Error, Ncount)

    elif type(metadata.dimension) == int and metadata.dimension != 0:

        original_xlabel = metadata.info["xlabel"]

        # All special characters are substituted with _ in McStas NeXus file
        x_field = re.sub(r'[^a-zA-Z]', "_", original_xlabel)

        if x_field not in f["entry1"]["data"][NeXus_field].keys():
            error_text = ("Didn't find xaxis in NeXus file. \n"
                          + "Expected this field for x axis: "
                          + str(x_field) + "\n"
                          + "Existing fields: "
                          + str(f["entry1"]["data"][NeXus_field].keys()))

            raise ValueError(error_text)

        xaxis = np.array(f["entry1"]["data"][NeXus_field][x_field])
        Intensity = np.array(f["entry1"]["data"][NeXus_field]["data"])
        Error = np.array(f["entry1"]["data"][NeXus_field]["errors"])
        Ncount = np.array(f["entry1"]["data"][NeXus_field]["ncount"])

        # The data is saved as a McStasDataBinned object
        return McStasDataBinned(metadata, Intensity, Error, Ncount, xaxis=xaxis)

    elif len(metadata.dimension) == 2:
        xaxis = []  # Assume evenly binned in 2d
        Intensity = np.array(f["entry1"]["data"][NeXus_field]["data"]).T
        Error = np.array(f["entry1"]["data"][NeXus_field]["errors"]).T
        Ncount = np.array(f["entry1"]["data"][NeXus_field]["ncount"]).T

        # The data is saved as a McStasDataBinned object
        return McStasDataBinned(metadata, Intensity, Error, Ncount, xaxis=xaxis)
    else:
        raise NameError(
            "Dimension not read correctly in data set "
            + "connected to monitor named "
            + metadata.component_name)


def load_monitor_text(metadata, data_folder_name):
    """
    Function that loads data given metadata and name of data folder
    This version is for a text file

    Loads data for single monitor and returns a McStasData object

    Parameters
    ----------

    metadata : McStasMetaData object
        McStasMetaData object corresponding to the monitor to be loaded

    data_folder_name : str
        path to folder from which metadata should be loaded
    """
    # Load data with numpy
    filename = os.path.join(data_folder_name, metadata.filename.rstrip())
    data = np.loadtxt(filename)

    # Split data into intensity, error and ncount
    if type(metadata.dimension) == int and metadata.dimension == 0:
        Intensity = data.T
        if metadata.total_E is None:
            Error = np.zeros(1)
        else:
            Error = np.array(metadata.total_E)

        if metadata.total_N is None:
            Ncount = np.zeros(1)
        else:
            Ncount = np.array(metadata.total_N)

        return McStasDataBinned(metadata, Intensity, Error, Ncount)

    elif type(metadata.dimension) == int and metadata.dimension != 0:
        xaxis = data.T[0, :]
        Intensity = data.T[1, :]
        Error = data.T[2, :]
        Ncount = data.T[3, :]

        # The data is saved as a McStasDataBinned object
        return McStasDataBinned(metadata, Intensity, Error, Ncount, xaxis=xaxis)

    elif len(metadata.dimension) == 2:
        # Need to check if it is binned data or event data

        with open(filename, 'rb', 0) as file, \
                mmap.mmap(file.fileno(), 0, access=mmap.ACCESS_READ) as s:
            if s.find(b'# Errors') != -1:
                data_type = "Binned"
            else:
                data_type = "Events"

        if data_type == "Events":
            Events = data

            return McStasDataEvent(metadata, Events)

        elif data_type == "Binned":
            # Binned 2D data
            xaxis = []  # Assume evenly binned in 2d
            data_lines = metadata.dimension[1]
            Intensity = data[0:data_lines, :]
            Error = data[data_lines:2 * data_lines, :]
            Ncount = data[2 * data_lines:3 * data_lines, :]

            # The data is saved as a McStasDataBinned object
            return McStasDataBinned(metadata, Intensity, Error, Ncount, xaxis=xaxis)
    else:
        raise NameError(
            "Dimension not read correctly in data set "
            + "connected to monitor named "
            + metadata.component_name)


def print_sim_output(sim_output):
    print(highlight(sim_output, "error", return_section=True, after_lines=10, highlight_type="FAIL"))
    print(highlight(sim_output, "error", return_section=False, highlight_type="FAIL"))


def highlight(string, search_term, return_section=False, highlight_type=None, after_lines=5):
    """
    Highlights search term in string and returns it, if return_section only sections with term is returned
    """

    search_term = search_term.lower()

    if not isinstance(string, str):
        return None

    # Early exit if search term is not in string
    output = string.lower().find(search_term)
    if output == -1:
        if return_section:
            return ""
        else:
            return string

    if return_section:
        instances = list(findall(string, search_term))
        n_instances = len(instances)
        print(f"---- Found {n_instances} places in McStas output with "
              f"keyword '{search_term}'. \n")

    if highlight_type is None:
        highlight_start = ""
        highlight_end = ""
    else:
        if not hasattr(bcolors, highlight_type):
            raise RuntimeError(f"Used highlight_type {highlight_type} "
                               f"in highlight not found in bcolors.")
        else:
            highlight_start = getattr(bcolors, highlight_type)
            highlight_end = bcolors.ENDC

    return_string = ""

    lines = string.split("\n")
    total_lines = len(lines)
    for index, line in enumerate(lines):
        output = line.lower().find(search_term)
        if output == -1:
            if not return_section:
                return_string += line + "\n"
        else:
            replaced_string = line[:output]
            replaced_string += highlight_start
            replaced_string += line[output:output + len(search_term)]
            replaced_string += highlight_end
            replaced_string += line[output + len(search_term):]
            replaced_string += "\n"

            return_string += replaced_string
            if return_section:
                extra_lines = min(total_lines - index, after_lines)
                for line_index in range(1, extra_lines):
                    line_to_include = lines[index + line_index]
                    if line_to_include.lower().find(search_term) != -1:
                        break
                    return_string += line_to_include + "\n"
                return_string += "-"*70 + "\n"

    return return_string


def findall(s, p):
    """
    Yields all the positions of the pattern p in the string s.
    """
    i = s.lower().find(p)
    while i != -1:
        yield i
        i = s.lower().find(p, i+1)


import os

from mcstasscript.interface.instr import McStas_instr
from mcstasscript.instr_reader.read_definition import DefinitionReader
from mcstasscript.instr_reader.read_declare import DeclareReader
from mcstasscript.instr_reader.read_uservars import UservarsReader
from mcstasscript.instr_reader.read_initialize import InitializeReader
from mcstasscript.instr_reader.read_trace import TraceReader
from mcstasscript.instr_reader.read_finally import FinallyReader


class InstrumentReader:
    """
    This class controls loading of a McStas file as a McStasScript object.

    This is done by reading the McStas file line by line while using the
    McStasScript API to load the information into the Instr object.

    Optionally a Python file with the McStasScript commands required to
    replicate the McStas instrument can be written to disk.

    Methods
    -------
    __init__(filename)
        Initializes reading of McStas instrument with given filename

    generate_py_version(product_filename)
        Generates a file named product_filename.py that recreates instr

    add_to_instr(Instr)
        Inserts information from instr file into McStasScript Instr instance
    """

    def __init__(self, filename):
        """
        Initialize the InstrumentReader with a target McStas instrument
        file. Use generate_py_version method for writing an eqvivalent
        McStasScript python file or the add_to_instr method to load this
        instrument file onto a Instr McStasScript object.
        """

        self.filename = filename
        self.Instr = None  # could set it up to create Instr
        self.write_file = False
        self.product_filename = "mc_script.py"
        self.instr_name = ""
        self.file_data = None
        self.line_index = 0
        self.file_length = 0

    def generate_py_version(self, product_filename):
        """
        Generate a McStasScript version of the instrument file used for
        initialize of the InstrumentReader object. The filename given is
        for the generated file.

        One should use this feature with some caution. Look through the
        generated McStasScript file and compare some output with the
        original insturment file to ensure everything was loaded correctly.
        """

        # Generate dummy instr object
        self.Instr = McStas_instr("dummy_object_for_generating_file")

        self.product_filename = product_filename
        self.write_file = True

        if os.path.isfile(self.product_filename):
            os.remove(self.product_filename)

        self._read_file()

    def add_to_instr(self, Instr):
        """
        Add contents of McStas instrument file selected in initialize
        to an McStasScript instrument object.
        """

        self.Instr = Instr
        self.write_file = False

        self._read_file()

    def _open_file(self):
        """
        Internal method that opens the instrument file to be read
        """

        with open(self.filename) as file:
            self.file_data = file.readlines()

        self.file_length = len(self.file_data)
        self.line_index = 0

    def _get_next_line(self):
        """
        Internalmethod that gets the next line to be read
        """

        line = self.file_data[self.line_index]
        self.line_index += 1
        return line

    def _return_line(self):
        """
        Internal method that puts line back into stack
        """

        self.line_index -= 1

    def _read_file(self):
        """
        Master method for reading the instrument file.  It goes through
        the file line by line, and checks which part of the instrument
        file it is currently reading. There are separate methods for
        reading the individual parts of the instrument file to reduce
        clutter.
        """

        # Initialize readers of the different McStas instrument sections
        args = [self.Instr, self.write_file, self.product_filename,
                self._get_next_line, self._return_line]
        self.Definition_reader = DefinitionReader(*args)
        self.Declare_reader = DeclareReader(*args)
        self.Uservars_reader = UservarsReader(*args)
        self.Initialize_reader = InitializeReader(*args)
        self.Trace_reader = TraceReader(*args)
        self.Finally_reader = FinallyReader(*args)

        # A mode for each type that activates the correct reader function
        definition_mode = False
        declare_mode = False
        uservars_mode = False
        initialize_mode = False
        trace_mode = False
        finally_mode = False
        comment_mode = False
        any_mode = False

        # check if instrument name has be read from file yet
        instr_name_read = False

        self._open_file()

        # for line in self.file_data:
        while self.line_index < self.file_length:

            line = self._get_next_line()

            # Find appropriate mode
            if line.strip().startswith("DEFINE INSTRUMENT") and not any_mode:
                definition_mode = True
                any_mode = True

            if line.strip().startswith("DECLARE") and not any_mode:
                declare_mode = True
                any_mode = True

            if line.strip().startswith("USERVARS") and not any_mode:
                uservars_mode = True
                any_mode = True

            if (line.strip().startswith("INITIALIZE") or
               line.strip().startswith("INITIALISE")) and not any_mode:
                initialize_mode = True
                any_mode = True

            if line.strip().startswith("TRACE") and not any_mode:
                trace_mode = True
                any_mode = True

            if line.strip().startswith("FINALLY") and not any_mode:
                finally_mode = True
                any_mode = True

            if line.strip().startswith("/*"):
                comment_mode = True

            # Read with appropriate reader
            if definition_mode and not comment_mode:
                # Get instrument name
                if not instr_name_read:
                    self.instr_name = line.split("(")[0].strip().split(" ")[-1]
                    instr_name_read = True
                    self.update_file_name()

                # Read line from definition
                definition_mode = self.Definition_reader.read_definition_line(line)
                # When read_definition finds the end, it will return False
                any_mode = definition_mode

            if declare_mode and not comment_mode:
                # Read line from declare
                declare_mode = self.Declare_reader.read_declare_line(line)
                # When read_declare finds the end, it will return False
                any_mode = declare_mode

            if uservars_mode and not comment_mode:
                # Read line from uservars
                uservars_mode = self.Uservars_reader.read_uservars_line(line)
                # When read_uservars finds the end, it will return False
                any_mode = uservars_mode

            if initialize_mode and not comment_mode:
                # Read line from initialize
                initialize_mode = self.Initialize_reader.read_initialize_line(line)
                # When read_initialize finds the end, it will return False
                any_mode = initialize_mode

            if trace_mode and not comment_mode:
                # Read line from initialize
                trace_mode = self.Trace_reader.read_trace_line(line)
                # When read_initialize finds the end, it will return False
                any_mode = trace_mode

            if finally_mode and not comment_mode:
                # Read line from finally
                finally_mode = self.Finally_reader.read_finally_line(line)
                # When read_finallyfinds the end, it will return False
                any_mode = finally_mode

            # Stop comment mode when end of comment block reached
            if "*/" in line.strip():
                comment_mode = False

    def update_file_name(self):
        """
        Updates filename for reader subclasses
        """

        self.Definition_reader.set_instr_name(self.instr_name)
        self.Declare_reader.set_instr_name(self.instr_name)
        self.Uservars_reader.set_instr_name(self.instr_name)
        self.Initialize_reader.set_instr_name(self.instr_name)
        self.Trace_reader.set_instr_name(self.instr_name)
        self.Finally_reader.set_instr_name(self.instr_name)



import unittest
import numpy as np
import copy

from mcstasscript.data.data import McStasDataBinned
from mcstasscript.data.data import McStasMetaData
from mcstasscript.jb_interface.simulation_interface import add_data

def set_dummy_MetaDataBinned_1d():
    """
    Sets up simple McStasMetaData object with dimension, 1d case
    """
    meta_data = McStasMetaData()
    meta_data.component_name = "component for 1d"
    meta_data.filename = "data.dat"
    meta_data.dimension = 50

    meta_data.info = {"Ncount" : 40}

    return meta_data


def set_dummy_McStasDataBinned_1d():
    """
    Sets up simple McStasData object, 1d case
    """
    meta_data = set_dummy_MetaDataBinned_1d()

    intensity = np.ones(20)
    error = np.ones(20)
    ncount = np.ones(20)
    axis = np.arange(20)*5.0

    return McStasDataBinned(meta_data, intensity, error, ncount, xaxis=axis)


def set_dummy_MetaDataBinned_2d():
    """
    Sets up simple McStasMetaData object with dimensions, 2d case
    """
    meta_data = McStasMetaData()
    meta_data.component_name = "test a component"
    meta_data.filename = "data.dat"
    meta_data.dimension = [50, 100]

    meta_data.info = {"Ncount": 40}

    return meta_data


def set_dummy_McStasDataBinned_2d():
    """
    Sets up simple McStasData object, 2d case
    """
    meta_data = set_dummy_MetaDataBinned_2d()

    intensity = np.ones(20).reshape(4, 5)
    error = np.ones(20).reshape(4, 5)
    ncount = np.ones(20).reshape(4, 5)

    return McStasDataBinned(meta_data, intensity, error, ncount)

class Test_add_data(unittest.TestCase):
    def test_1d_updates_correctly(self):
        """
        Test that adding 1d dataset modifies only the intended dataset
        """

        data1 = set_dummy_McStasDataBinned_1d()
        data1_original = copy.deepcopy(data1)

        data2 = set_dummy_McStasDataBinned_1d()
        data2_original = copy.deepcopy(data2)

        add_data([data1], [data2])

        # Data 2 should not be touched
        self.assertTrue(np.array_equal(data2.Intensity, data2_original.Intensity))
        self.assertTrue(np.array_equal(data2.Error, data2_original.Error))
        self.assertTrue(np.array_equal(data2.Ncount, data2_original.Ncount))

        # Data 1 Intensity should be unchanged, as data1 and data2 equal
        self.assertTrue(np.array_equal(data1.Intensity, data1_original.Intensity))
        # Data 1 should be updated
        self.assertFalse(np.array_equal(data1.Error, data1_original.Error))
        self.assertFalse(np.array_equal(data1.Ncount, data1_original.Ncount))

    def test_1d_updates_different(self):
        """
        Test that adding 1d datasets work as expected when different
        """
        data1 = set_dummy_McStasDataBinned_1d()
        data1.Intensity *= 2.0
        data1.Intensity[10:] *= 2.0
        data1.Error *= 1.5
        data1.Ncount *= 4.0
        data1.metadata.info["Ncount"] *= 4.0
        data1_original = copy.deepcopy(data1)

        data2 = set_dummy_McStasDataBinned_1d()
        data2.Intensity *= 3.0
        data2.Error *= 1.5
        data2_original = copy.deepcopy(data2)

        add_data([data1], [data2])

        # Data 2 should not be touched
        self.assertTrue(np.array_equal(data2.Intensity, data2_original.Intensity))
        self.assertTrue(np.array_equal(data2.Error, data2_original.Error))
        self.assertTrue(np.array_equal(data2.Ncount, data2_original.Ncount))

        # 4 times more weight on data1, intensity 2 and 3
        expected_low_intensity = 4/5*2.0 + 1/5*3.0
        # 4 times more weight on data1, intensity 4 and 3
        expected_high_intensity = 4/5*2.0*2.0 + 1/5*3.0
        expected_error = np.sqrt((4 / 5) ** 2 * 1.5 ** 2 + (1 / 5) ** 2 * 1.5 ** 2)

        for index in range(len(data1_original.Intensity)):
            if index < 10:
                self.assertEqual(data1.Intensity[index], expected_low_intensity)
            else:
                self.assertEqual(data1.Intensity[index], expected_high_intensity)

            self.assertEqual(data1.Error[index], expected_error)
            self.assertEqual(data1.Ncount[index], 5.0)

        self.assertEqual(data1.metadata.info["Ncount"], 40*4+40)

    def test_fail(self):
        """
        Test that adding datasets fail when they dont have the same monitors

        Both 1d and 2d cases included.
        """

        data11 = set_dummy_McStasDataBinned_1d()
        data11.name = "first monitor"
        data11.filename = "first_monitor.dat"
        data12 = set_dummy_McStasDataBinned_2d()
        data12.name = "second monitor"
        data12.filename = "second_monitor.dat"
        data13 = set_dummy_McStasDataBinned_1d()
        data13.name = "third monitor"
        data13.filename = "third_monitor.dat"

        data21 = set_dummy_McStasDataBinned_1d()
        data21.name = "first monitor"
        data21.filename = "first_monitor.dat"
        data22 = set_dummy_McStasDataBinned_2d()
        data22.name = "second monitor"
        data22.filename = "second_monitor.dat"
        data23 = set_dummy_McStasDataBinned_1d()
        data23.name = "third monitor"
        data23.filename = "third_monitor.dat"

        # Should succeed, monitors match
        add_data([data11, data12, data13], [data21, data22, data23])
        # Should succeed, all monitors needed to update first argument present
        add_data([data11, data12], [data21, data22, data23])

        # Should fail if a monitor is missing
        with self.assertRaises(NameError):
            add_data([data11, data12, data13], [data21, data22])

        data23.name = "different monitor"
        # Should fail if name mismatch
        with self.assertRaises(NameError):
            add_data([data11, data12, data13], [data21, data22, data23])

    def test_2d_updates_correctly(self):
        """
        Test that adding 1d dataset modifies only the intended dataset
        """

        data1 = set_dummy_McStasDataBinned_2d()
        data1_original = copy.deepcopy(data1)

        data2 = set_dummy_McStasDataBinned_2d()
        data2_original = copy.deepcopy(data2)

        add_data([data1], [data2])

        # Data 2 should not be touched
        self.assertTrue(np.array_equal(data2.Intensity, data2_original.Intensity))
        self.assertTrue(np.array_equal(data2.Error, data2_original.Error))
        self.assertTrue(np.array_equal(data2.Ncount, data2_original.Ncount))

        # Data 1 Intensity should be unchanged, as data1 and data2 equal
        self.assertTrue(np.array_equal(data1.Intensity, data1_original.Intensity))
        # Data 1 should be updated
        self.assertFalse(np.array_equal(data1.Error, data1_original.Error))
        self.assertFalse(np.array_equal(data1.Ncount, data1_original.Ncount))

    def test_2d_updates_different(self):
        """
        Test that adding 2d datasets work as expected when different
        """
        data1 = set_dummy_McStasDataBinned_2d()
        data1.Intensity *= 2.0
        data1.Intensity[1,:] *= 2.0
        data1.Error *= 1.5
        data1.Ncount *= 4.0
        data1.metadata.info["Ncount"] *= 4.0
        data1_original = copy.deepcopy(data1)

        data2 = set_dummy_McStasDataBinned_2d()
        data2.Intensity *= 3.0
        data2.Error *= 1.5
        data2_original = copy.deepcopy(data2)

        add_data([data1], [data2])

        # Data 2 should not be touched
        self.assertTrue(np.array_equal(data2.Intensity, data2_original.Intensity))
        self.assertTrue(np.array_equal(data2.Error, data2_original.Error))
        self.assertTrue(np.array_equal(data2.Ncount, data2_original.Ncount))

        # 4 times more weight on data1, intensity 2 and 3
        expected_low_intensity = 4/5*2.0 + 1/5*3.0
        # 4 times more weight on data1, intensity 4 and 3
        expected_high_intensity = 4/5*2.0*2.0 + 1/5*3.0
        expected_error = np.sqrt((4 / 5) ** 2 * 1.5 ** 2 + (1 / 5) ** 2 * 1.5 ** 2)

        for index1 in range(len(data1_original.Intensity[:,0])):
            for index2 in range(len(data1_original.Intensity[0, :])):

                if index1 == 1:
                    self.assertEqual(data1.Intensity[index1, index2], expected_high_intensity)
                else:
                    self.assertEqual(data1.Intensity[index1, index2], expected_low_intensity)

            self.assertEqual(data1.Error[index1, index2], expected_error)
            self.assertEqual(data1.Ncount[index1, index2], 5.0)

        self.assertEqual(data1.metadata.info["Ncount"], 40*4+40)




import copy

import numpy as np
import matplotlib
import matplotlib.pyplot as plt

from matplotlib.ticker import MaxNLocator
from matplotlib.colors import BoundaryNorm
import matplotlib.ticker as mticker

from mcstasscript.data.data import McStasData
from mcstasscript.data.data import McStasDataEvent


def remove_eventdata(data_list, verbose=True):
    """
    Removes event data from a list, useful as these can't be plotted
    """
    reduced_data_list = []
    skipped_names = []
    for element in data_list:
        if not isinstance(element, McStasDataEvent):
            reduced_data_list.append(element)
        else:
            skipped_names.append(element.metadata.component_name)

    if verbose:
        for name in skipped_names:
            print(f"Skipped plotting {name} as it contains event data.")

    return reduced_data_list


def _fmt(x, pos):
    """
    Used for nice formatting of powers of 10 when plotting logarithmic
    """
    a, b = '{:.2e}'.format(x).split('e')
    b = int(b)
    if abs(float(a) - 1) < 0.01:
        return r'$10^{{{}}}$'.format(b)
    else:
        return r'${}\cdot 10^{{{}}}$'.format(a, b)


def _fmt(x, pos):
    a, b = f"{x:.2e}".split("e")
    b = int(b)

    if abs(float(a) - 1) < 0.01:
        return rf"10^{{{b}}}"
    else:
        return rf"{a}\cdot 10^{{{b}}}"

formatter = mticker.FuncFormatter(lambda x, pos: rf"${_fmt(x, pos)}$")


def _find_min_max_I(data):
    """
    Returns minimum and maximum intensity to plot given dataset

    Uses the plot options embedded in McStasData to determine the proper
    minimum and maximum intensity to display in a plot.

    Have to take cut_min and cut_max into account that can cut parts of
    the intensity away. When plotting logarithmic, orders_of_mags limits
    the orders of magnitude shown.

    Returns tuple of minimum and maximum, when no data is present the
    function returns 0, 0.
    """
    cut_max = data.plot_options.cut_max  # Default 1
    cut_min = data.plot_options.cut_min  # Default 0

    to_plot = data.Intensity

    min_value = to_plot.min()
    max_value = to_plot.max()

    if min_value == 0 and max_value == 0:
        return 0, 0

    if not data.plot_options.log:
        # Linear, simple case
        # Cut top and bottom of data as specified in cut variables
        min_value = min_value + (max_value - min_value) * cut_min
        max_value = max_value * cut_max

    else:
        # Logarithmic, minimum / maximum can not be zero
        max_data_value = to_plot.max()
        max_value = np.log10(max_data_value * cut_max)

        min_value = np.min(to_plot[np.nonzero(to_plot)])
        min_value = min_value + (max_data_value - min_value) * cut_min
        min_value = np.log10(min_value)

        # Take orders_of_mag into account (max / min in log10)
        if max_value - min_value > data.plot_options.orders_of_mag:
            min_value = max_value - data.plot_options.orders_of_mag

        # Convert back from log10
        min_value = 10.0 ** min_value
        max_value = 10.0 ** max_value

    return min_value, max_value


def _plot_fig_ax(data, fig, ax, **kwargs):
    """
    Plots the content of a single McStasData object

    Plotting is controlled through options associated with the
    McStasData objects.

    When plotting 2D objects, returns the pcolormesh object
    """

    if type(data.metadata.dimension) == int and data.metadata.dimension == 0:
        # Can't plot 0D data, show the info
        ax.text(0.5, 0.9, data.metadata.title, ha="center")
        ax.text(0.5, 0.7, "I: " + str(float(data.Intensity)), ha="center")
        ax.text(0.5, 0.5, "E: " + str(float(data.Error)), ha="center")
        ax.text(0.5, 0.3, "N: " + str(int(data.Ncount)), ha="center")
        ax.axis("off")

    elif type(data.metadata.dimension) == int and data.metadata.dimension != 0:

        x_axis_mult = data.plot_options.x_limit_multiplier

        x = data.xaxis * x_axis_mult
        y = data.Intensity
        y_err = data.Error

        ax.errorbar(x, y, yerr=y_err)

        ax.set_xlim(data.metadata.limits[0] * x_axis_mult,
                    data.metadata.limits[1] * x_axis_mult)

        # Add a title
        ax.set_title(data.metadata.title)

        # Add axis labels
        ax.set_xlabel(data.metadata.xlabel)
        ax.set_ylabel(data.metadata.ylabel)

        if data.plot_options.custom_xlim_left:
            ax.set_xlim(left=data.plot_options.left_lim)

        if data.plot_options.custom_xlim_right:
            ax.set_xlim(right=data.plot_options.right_lim)

        if data.plot_options.log:
            ax.set_yscale("log", nonpositive='clip')

            n_non_zero = np.count_nonzero(data.Intensity)
            if n_non_zero == 0:
                # Plot is empty, return
                return

            non_zero = np.nonzero(data.Intensity)
            min_value_log = np.log10(min(data.Intensity[non_zero]))
            max_value_log = np.log10(max(data.Intensity[non_zero]))

            orders_of_mag = data.plot_options.orders_of_mag
            if max_value_log - min_value_log > orders_of_mag:
                ax.set_ylim(top=1.1*10.0 ** max_value_log)
                ax.set_ylim(bottom=10.0 ** (max_value_log - orders_of_mag))

    elif len(data.metadata.dimension) == 2:

        min_value, max_value = _find_min_max_I(data)

        if "fixed_minimum_value" in kwargs:
            min_value = kwargs["fixed_minimum_value"]
        if "fixed_maximum_value" in kwargs:
            max_value = kwargs["fixed_maximum_value"]

        # Set the axis
        x_axis_mult = data.plot_options.x_limit_multiplier
        y_axis_mult = data.plot_options.y_limit_multiplier

        X = np.linspace(data.metadata.limits[0] * x_axis_mult,
                        data.metadata.limits[1] * x_axis_mult,
                        data.metadata.dimension[0] + 1)
        Y = np.linspace(data.metadata.limits[2] * y_axis_mult,
                        data.metadata.limits[3] * y_axis_mult,
                        data.metadata.dimension[1] + 1)

        # Create a meshgrid for both x and y
        x, y = np.meshgrid(X, Y)

        # Generate information on necessary colorrange
        levels = MaxNLocator(nbins=150).tick_values(min_value, max_value)

        # Select colormap
        cmap = copy.copy(plt.get_cmap(data.plot_options.colormap))
        if "no_data_to_black" in kwargs:
            if kwargs["no_data_to_black"]:
                cmap.set_bad((0, 0, 0))

        norm = BoundaryNorm(levels, ncolors=cmap.N, clip=True)

        # Empty data, return without cmap or norm
        if min_value == 0 and max_value == 0:
            levels = MaxNLocator(nbins=150).tick_values(0.001, 1.0)
            norm = BoundaryNorm(levels, ncolors=cmap.N, clip=True)
            im = ax.pcolormesh(x, y, data.Intensity, cmap=cmap, norm=norm)

        # Plot the data on the meshgrids
        elif data.plot_options.log:
            color_norm = matplotlib.colors.LogNorm(vmin=min_value,
                                                   vmax=max_value)
            im = ax.pcolormesh(x, y, data.Intensity,
                               cmap=cmap, norm=color_norm)
        else:
            im = ax.pcolormesh(x, y, data.Intensity, cmap=cmap, norm=norm)

        # Add the colorbar
        if data.plot_options.show_colorbar:
            cax = None
            if "colorbar_axes" in kwargs:
                cax = kwargs["colorbar_axes"]

            colorbar = fig.colorbar(im, ax=ax, cax=cax)

            # If your colorbar is logarithmic:
            #colorbar.locator = mticker.LogLocator(base=10)
            #colorbar.formatter = mticker.LogFormatterMathtext(base=10)

            #colorbar.update_ticks()

            if data.metadata.zlabel is not None:
                colorbar.set_label(data.metadata.zlabel)

            if "colorbar_axes" in kwargs:
                cax.set_aspect(20)

        # Add a title
        ax.set_title(data.metadata.title)

        # Add axis labels
        ax.set_xlabel(data.metadata.xlabel)
        ax.set_ylabel(data.metadata.ylabel)

        if data.plot_options.custom_ylim_top:
            ax.set_ylim(top=data.plot_options.top_lim)

        if data.plot_options.custom_ylim_bottom:
            ax.set_ylim(bottom=data.plot_options.bottom_lim)

        if data.plot_options.custom_xlim_left:
            ax.set_xlim(left=data.plot_options.left_lim)

        if data.plot_options.custom_xlim_right:
            ax.set_xlim(right=data.plot_options.right_lim)

        return im
    else:
        print("Error, dimension not read correctly")


def _handle_kwargs(data_list, **kwargs):
    """
    Handle kwargs when list of McStasData objects given.

    Returns data_list

    data_list is turned into a list if it isn't already
    event data is removed as it can't be plotted directly

    Any kwargs can be given as a list, in that case apply them to given
    to the corresponding index.
    """

    if "fontsize" in kwargs:
        used_fontsize = kwargs["fontsize"]
    else:
        used_fontsize = 11
    plt.rcParams.update({'font.size': used_fontsize})

    if isinstance(data_list, McStasData):
        # Only a single element, put it in a list for easier syntax later
        data_list = [data_list]

    # Remove event data that can't be plotted in meaningful way
    data_list = remove_eventdata(data_list)

    known_plotting_kwargs = ["log", "orders_of_mag",
                             "top_lim", "bottom_lim",
                             "left_lim", "right_lim",
                             "cut_min", "cut_max",
                             "colormap", "show_colorbar",
                             "x_axis_multiplier",
                             "y_axis_multiplier"]

    for option in known_plotting_kwargs:
        if option in kwargs:
            given_option = kwargs[option]

            if isinstance(given_option, list):
                if len(data_list) < len(given_option):
                    raise ValueError("Keyword argument " + option + " is "
                                     + "given as a list, but this list has "
                                     + "more elements than there are "
                                     + "data sets to be plotted.")

                index = 0
                for per_list_option in given_option:
                    input_kwarg = {option: per_list_option}
                    data_list[index].set_plot_options(**input_kwarg)
                    index += 1

            else:
                for data in data_list:
                    input_kwarg = {option: given_option}
                    data.set_plot_options(**input_kwarg)

            # Remove option from kwargs
            del kwargs[option]

    return data_list


import numpy as np

class Layer:
    def __init__(self, name, cryostat,
                 origin_to_bottom, bottom_thickness, origin_to_top, top_thickness=0.01,
                 inner_radius=None, outer_radius=None, thickness=None,
                 material="Al", p_interact=0):
        """
        Layer of cryostat with a shell made of given material and hollowed by vacuum

        Describes a layer of a cryostat with given geometry and materials. The geometry
        is specified with distance from origin to top and bottom, along with a
        thickness of these. In addition an outer and inner radius is needed, here the
        user can specify two of inner_radius, outer_radius and thickness.
        The Layer is given the cryostat object of which is it a part to access methods
        and attributes in a simple way.
        The layer is built from Union components, and so are not allowed to perfectly
        intersect the planes from other layers. These are registered and check with the
        cryostat class. If a custom material is given instead of Al, that material has
        to exist as a Union_make_material component.

        It is possible to add windows to a layer, these are added with the add_window
        method, windows can be a reduction of radius from the inside, outside or both.

        Parameters
        ----------

        origin_to_bottom: float
            Distance from origin to bottom of layer [m] (bottom thickness added)

        bottom_thickness: float
            Thickness of layer bottom [m]

        origin_to_top: float
            Distance from origin to top of layer [m] (top thickness added)

        top_thickness: float
            Thickness of layer top [m]

        inner_radius: float
            Inner radius of the cryostat layer [m]

        outer_radius: float
            Outer radius of the cryostat layer [m]

        thickness: float
            Thickness of the shell, can be provided instead of either inner or outer radius [m]

        material: string
            Material of which the cryostat layer should be made

        p_interact: float
            p_interact set for all Union geometry components
        """

        self.name = name
        self.material = material
        self.p_interact = p_interact

        self.cryostat = cryostat
        self.instr = self.cryostat.instr
        self.reference = self.cryostat.origin

        try:
            self.instr.get_component(self.material)
        except:
            raise RuntimeError("No material named '" + str(self.material)
                               + "' found in given instrument file. \n"
                               + "Construct with Union_make_material.")

        self.shell_lowest_point = -origin_to_bottom - bottom_thickness
        self.shell_highest_point = origin_to_top + top_thickness
        self.shell_center = 0.5 * (self.shell_lowest_point + self.shell_highest_point)
        self.shell_height = self.shell_highest_point - self.shell_lowest_point

        # Check with the database these wont cause a collision
        self.cryostat.add_y_plane(self.shell_lowest_point)
        self.cryostat.add_y_plane(self.shell_highest_point)

        self.vac_lowest_point = -origin_to_bottom
        self.vac_highest_point = origin_to_top
        self.vac_center = 0.5 * (self.vac_highest_point + self.vac_lowest_point)
        self.vac_height = self.vac_highest_point - self.vac_lowest_point

        # Check with the database these wont cause a collision
        self.cryostat.add_y_plane(self.vac_lowest_point)
        self.cryostat.add_y_plane(self.vac_highest_point)

        self.inner_radius, self.outer_radius = self.check_radius_input(inner_radius, outer_radius, thickness)

        # Check with the database these wont cause a collision
        self.cryostat.add_radius(self.inner_radius)
        self.cryostat.add_radius(self.outer_radius)

        # Set up shell
        layer = self.instr.add_component(self.name + "_layer", "Union_cylinder")
        layer.material_string = '"' + self.material + '"'
        layer.p_interact = self.p_interact

        layer.radius = self.outer_radius
        layer.yheight = self.shell_height

        layer.set_AT([0, self.shell_center, 0], RELATIVE=self.reference)

        self.layer = layer

        # Set up vacuum
        layer_vac = self.instr.add_component(self.name + "_layer_vac", "Union_cylinder")
        layer_vac.material_string = '"Vacuum"'

        layer_vac.radius = self.inner_radius
        layer_vac.yheight = self.vac_height

        layer_vac.set_AT([0, self.vac_center, 0], RELATIVE=self.reference)

        self.layer_vac = layer_vac

        # List of union components, highest priority first
        self.outer_cuts = []
        self.inner_cuts = []
        # self.main_union_components = [self.layer_vac, self.layer]
        self.union_components = [self.layer_vac, self.layer]

        # Prepare for windows
        self.inner_window_index = 0
        self.outer_window_index = 0

    def count_inputs(self, *args):
        """
        Counts how many of given inputs are not None
        """
        n_inputs = 0
        for arg in args:
            if arg is not None:
                n_inputs += 1
        return n_inputs

    def check_radius_input(self, inner_radius, outer_radius, thickness):
        """
        Returns inner and outer radius from user input that may include thickness

        The inner_radius, outer_radius and thickness can have one not specified,
        and this method will calculate the inner and out radius from the two given.
        A input not given is specified as None.
        """

        n_inputs = self.count_inputs(inner_radius, outer_radius, thickness)

        if n_inputs != 2:
            raise RuntimeError("Set two of inner_radius, outer_radius, and thickness.")

        if inner_radius is None:
            inner_radius = outer_radius - thickness
        else:
            inner_radius = inner_radius

        if outer_radius is None:
            outer_radius = inner_radius + thickness
        else:
            outer_radius = outer_radius

        return inner_radius, outer_radius

    def add_window(self, inner_radius=None, outer_radius=None, thickness=None,
                   height=None, origin_to_top=None, origin_to_bottom=None):
        """
        Adds 360 deg window in given height interval

        Adds window as a reduction in thickness of the layer in a certain height
        interval. The reduction can be both from the outside, inside or both.
        If only outer_radius is given, a reduction of the outside is assumed.
        If only inner_radius is given, the window will be on the inside.
        If two of the three parameters inner_radius, outer_radius and thickness
        is given, material will be removed from both the outside and inside, only
        use this if necessary, never specify inner / outer radius identical to the
        main layer.
        Any two of height, origin_to_top and origin_to_bottom can be given, or if
        the window is symmetrical in height around the origin position, the height
        alone is sufficient.
        It is possible to add multiple multiple windows by calling this method
        multiple times, add the tallest first.

        Keyword arguments
        -----------------

        inner_radius : float
            inner radius of window (should be larger than main inner_radius)

        outer_radius : float
            outer radius of window (should be less than main outer_radius)

        thickness : float
            thickness of the window

        height : float
            height of window

        origin_to_top : float
            distance from origin to top of window

        origin_to_bototm : float
            distance from origin to bottom of window
        """

        # Assume no cuts are made
        outer_cut = False
        inner_cut = False

        # Check if cuts are made to inner, outer or both
        if self.count_inputs(inner_radius, outer_radius, thickness) == 1:
            if inner_radius is not None:
                inner_cut = True
                outer_radius = self.outer_radius
            elif outer_radius is not None:
                outer_cut = True
                inner_radius = self.inner_radius
            else:
                # could center this window
                raise RuntimeError("Cant tell if window is on inside or outside.")

        inner_radius, outer_radius = self.check_radius_input(inner_radius, outer_radius, thickness)

        if inner_radius < self.inner_radius:
            raise RuntimeError("Window has smaller inner radius than main layer, needs to be larger.")

        if outer_radius > self.outer_radius:
            raise RuntimeError("Window has larger outer radius than main layer, needs to be smaller.")

        if abs(inner_radius - self.inner_radius) > 1E-5:
            inner_cut = True

        if abs(outer_radius - self.outer_radius) > 1E-5:
            outer_cut = True

        if self.count_inputs(height, origin_to_top, origin_to_bottom) == 3:
            raise RuntimeError("Ambigious definition of window height.")

        if height is not None:
            window_top = height / 2
            window_bottom = -height / 2
            if origin_to_top is not None:
                window_top = origin_to_top
                window_bottom = origin_to_top - height
            if origin_to_bottom is not None:
                window_top = -origin_to_bottom + height
                window_bottom = -origin_to_bottom
        else:
            window_top = origin_to_top
            window_bottom = -origin_to_bottom

        window_height = window_top - window_bottom
        window_position = 0.5 * (window_top + window_bottom)

        if outer_cut:
            name = self.name + "_outer_cut_" + str(self.outer_window_index)
            o_cut = self.instr.add_component(name, "Union_cylinder")
            o_cut.set_AT([0, window_position, 0], RELATIVE=self.reference)

            o_cut.material_string = '"Vacuum"'
            o_cut.radius = self.outer_radius + 1E-6
            o_cut.yheight = window_height + 1E-6

            # Check these radius and y_planes do not collide with others
            self.cryostat.add_radius(o_cut.radius)
            self.cryostat.add_y_plane(window_top + 0.5E-6)
            self.cryostat.add_y_plane(window_bottom - 0.5E-6)

            name = self.name + "_outer_cut_replace_" + str(self.outer_window_index)
            o_cut_m = self.instr.add_component(name, "Union_cylinder")
            o_cut_m.set_AT([0, window_position, 0], RELATIVE=self.reference)

            o_cut_m.material_string = '"' + self.material + '"'
            o_cut_m.p_interact = self.p_interact
            o_cut_m.radius = outer_radius
            o_cut_m.yheight = window_height + 5E-6

            # Check these radius and y_planes do not collide with others
            self.cryostat.add_radius(o_cut_m.radius)
            self.cryostat.add_y_plane(window_top + 2.5E-6)
            self.cryostat.add_y_plane(window_bottom - 2.5E-6)

            # Update list of outer cut components, order is important
            self.outer_cuts = [o_cut_m, o_cut] + self.outer_cuts

            self.outer_window_index += 1

        if inner_cut:
            name = self.name + "_inner_cut_" + str(self.inner_window_index)
            i_cut = self.instr.add_component(name, "Union_cylinder")
            i_cut.set_AT([0, window_position, 0], RELATIVE=self.reference)

            i_cut.material_string = '"Vacuum"'
            i_cut.radius = inner_radius
            i_cut.yheight = window_height

            # Check these radius and y_planes do not collide with others
            self.cryostat.add_radius(i_cut.radius)
            self.cryostat.add_y_plane(window_top)
            self.cryostat.add_y_plane(window_bottom)

            # Update list of inner cut components, order is important
            self.inner_cuts = [i_cut] + self.inner_cuts

            self.inner_window_index += 1

        # Create full list of Union components used in order from highest to lowest priority
        self.union_components = [self.layer_vac] + self.inner_cuts + self.outer_cuts + [self.layer]


class Cryostat:
    def __init__(self, name, instr, reference="PREVIOUS",
                 min_priority=20, max_priority=100):
        """
        Handles addition of a cryostat with multiple layers and windows

        This class can add a description of a cryostat to a McStasScript
        instrument object. The cryostat is made of several layers consisting
        of a shell and vacuum, these are added with the add_layer method and
        should be added in the order from smallest (inside) to largest. Each
        layer can be accessed with the last_layer attribute, and one can add
        windows to each layer. Consult the Layer class for details. When all
        layers are added, it is necessary to build the cryostat, this step is
        necessary to adjust the priority of each Union component.
        The position of the cryostat can be adjusted with the set_AT and
        set_ROTATED methods that works as the standard McStasScript versions.
        It is possible to add logger components that show scattering in the
        system with the add_spatial_loggers method.
        The position of the cryostat refers to the sample position and is
        called the origin. It can be placed in McStas with set_AT and
        set_ROTATED methods on the cryostat object.

        Parameters
        ----------

        name : str
            Name of the cryostat, will be used in naming of all added components

        instr : instrument object inherited from McCode_instr
            Instrument object where the cryostat should be added

        Keyword arguments
        -----------------

        reference : str
            Name of the component which the cryostat should be located relative to

        min_priority : float
            Minimum Union priority used (default 20)

        max_priority : float
            Maximum Union priority used, add a sample with higher priority (default 100)
        """
        self.name = str(name)
        self.instr = instr
        self.reference = reference
        self.min_priority = min_priority
        self.max_priority = max_priority

        self.layers = []
        self.last_layer = None

        self.origin = self.instr.add_component(self.name, "Arm")
        self.origin.set_AT([0, 0, 0], RELATIVE=reference)

        # Check if Al exists, if not add it!
        try:
            Al_component = None
            Al_component = instr.get_component("Al")
        except:
            Al_inc = self.instr.add_component("Al_inc", "Incoherent_process")
            Al_inc.sigma = 4 * 0.0082  # Incoherent cross section in Barns
            Al_inc.unit_cell_volume = 66.4  # Unit cell volume in AA^3

            Al_pow = self.instr.add_component("Al_pow", "Powder_process")
            Al_pow.reflections = "\"Al.laz\""  # Data file with powder lines

            Al = self.instr.add_component("Al", "Union_make_material")
            Al.my_absorption = "100*4*0.231/66.4"  # Inverse penetration depth in 1/m
            Al.process_string = '"Al_inc,Al_pow"'  # Make a material with aluminium incoherent and aluminium powder`

        # Set intial state of loggers, not having been added yet
        # Important to avoid multiple sets of the same kind added
        self.spatial_loggers_set = False
        self.time_logger_set = False
        self.animation_loggers = []

        # Height database
        self.used_y_planes = []
        self.used_radius_values = []

    def set_AT(self, *args, **kwargs):
        """
        Sets position of cryostat, sample position used as reference
        """
        self.origin.set_AT(*args, **kwargs)

    def set_ROTATED(self, *args, **kwargs):
        """
        Sets rotation of cryostat, sample position used as reference
        """
        self.origin.set_AT(*args, **kwargs)

    def add_layer(self, *args, **kwargs):
        """
        Adds layer to cryostat, all arguments passed to Layer. Consult Layer
        class for additional help on adding a layer.
        """

        if "name" not in kwargs:
            kwargs["name"] = self.name + "_layer_" + str(len(self.layers))

        if "cryostat" not in kwargs:
            kwargs["cryostat"] = self

        layer = Layer(*args, **kwargs)
        self.last_layer = layer
        self.layers.append(layer)

    def find_cryostat_dimensions(self):
        """
        Returns spatial extend of cryostat with padding for plotting

        Returns tuple with x, lowest y, highest y and z
        """
        # find outer dimensions:
        max_radius = self.last_layer.outer_radius
        highest_point = self.last_layer.shell_highest_point
        lowest_point = self.last_layer.shell_lowest_point

        if highest_point > abs(lowest_point):
            height = 2 * highest_point
        else:
            height = 2 * abs(lowest_point)

        mon_z = 1.1 * max_radius
        mon_x = 1.1 * max_radius
        if lowest_point < 0:
            mon_y_low = 1.1 * lowest_point
        else:
            mon_y_low = 0.9 * lowest_point

        if highest_point > 0:
            mon_y_high = 1.1 * highest_point
        else:
            mon_y_high = 0.9 * highest_point

        return mon_x, mon_y_low, mon_y_high, mon_z

    def add_spatial_loggers(self, n_x=500, n_y=500, n_z=500):
        """
        Adds spatial Union loggers to the code

        The spatial loggers will be set so they cover the entire cryostat and
        include views from top, side and front along with a close up of the
        windows as a slice in zy with limited z of +/- 5 mm.

        Keyword arguments
        -----------------

        n_x : int
            Number of bins in x direction

        n_y : int
            Number of bins in y direction

        n_z : int
            Number of bins in z direction
        """

        if self.spatial_loggers_set:
            raise RuntimeError("Can not add two sets of spatial loggers.")

        self.spatial_loggers_set = True

        # find outer dimensions:
        mon_x, mon_y_low, mon_y_high, mon_z = self.find_cryostat_dimensions()

        space_2D_zx = self.instr.add_component(self.name + "_logger_space_zx", "Union_logger_2D_space")
        space_2D_zx.set_AT([0, 0, 0], RELATIVE=self.origin)
        space_2D_zx.filename = '"' + self.name + '_space_zx.dat"'
        space_2D_zx.D_direction_1 = '"z"'
        space_2D_zx.n1 = n_z
        space_2D_zx.D1_min = -mon_z
        space_2D_zx.D1_max = mon_z
        space_2D_zx.D_direction_2 = '"x"'
        space_2D_zx.n2 = n_x
        space_2D_zx.D2_min = -mon_x
        space_2D_zx.D2_max = mon_x

        space_2D_zy = self.instr.add_component(self.name + "_logger_space_zy", "Union_logger_2D_space")
        space_2D_zy.set_AT([0, 0, 0], RELATIVE=self.origin)
        space_2D_zy.filename = '"' + self.name + '_space_zy.dat"'
        space_2D_zy.D_direction_1 = '"z"'
        space_2D_zy.n1 = n_z
        space_2D_zy.D1_min = -mon_z
        space_2D_zy.D1_max = mon_z
        space_2D_zy.D_direction_2 = '"y"'
        space_2D_zy.n2 = n_y
        space_2D_zy.D2_min = mon_y_low
        space_2D_zy.D2_max = mon_y_high

        space_2D_xy = self.instr.add_component(self.name + "_logger_space_xy", "Union_logger_2D_space")
        space_2D_xy.set_AT([0, 0, 0], RELATIVE=self.origin)
        space_2D_xy.filename = '"' + self.name + '_space_xy.dat"'
        space_2D_xy.D_direction_1 = '"x"'
        space_2D_xy.n1 = n_x
        space_2D_xy.D1_min = -mon_x
        space_2D_xy.D1_max = mon_x
        space_2D_xy.D_direction_2 = '"y"'
        space_2D_xy.n2 = n_y
        space_2D_xy.D2_min = mon_y_low
        space_2D_xy.D2_max = mon_y_high

        # Adding monitor that shows windows better
        lowest_inner_radius = self.layers[0].inner_radius
        largest_outer_radius = self.layers[-1].outer_radius

        space_3D_zy = self.instr.add_component(self.name + "_logger_space_zy_close", "Union_logger_3D_space")
        space_3D_zy.set_AT([0, 0, 0], RELATIVE=self.origin)
        space_3D_zy.filename = '"' + self.name + '_space_zy_close.dat"'
        space_3D_zy.D_direction_1 = '"z"'
        space_3D_zy.n1 = n_z
        space_3D_zy.D1_min = -1.02 * largest_outer_radius
        space_3D_zy.D1_max = -0.9 * lowest_inner_radius
        space_3D_zy.D_direction_2 = '"y"'
        space_3D_zy.n2 = n_y
        space_3D_zy.D2_min = mon_y_low
        space_3D_zy.D2_max = mon_y_high
        space_3D_zy.D_direction_3 = '"x"'
        space_3D_zy.n3 = 1
        space_3D_zy.D3_min = -0.005
        space_3D_zy.D3_max = 0.005

    def add_time_histogram(self, t_min=0, t_max=0.1):
        """
        Adds histogram of scattering intensity as function of time

        Very useful when setting time range for animation, only one can be added.

        Parameters
        ----------

        t_min : float
            Lowest time recorded in [s]

        t_max : float
            Highest time recorded in [s]
        """

        if self.time_logger_set:
            raise RuntimeError("Can not add two sets of time_histogram loggers.")

        self.time_logger_set = True

        time_mon = self.instr.add_component(self.name + "_logger_time", "Union_logger_1D")
        time_mon.variable = '"time"'
        time_mon.min_value = t_min
        time_mon.max_value = t_max
        time_mon.n1 = 1000

    def add_animation(self, t_min=0, t_max=0.1, n_frames=10,
                      d1="z", n1=300, d2="y", n2=300):
        """
        Adds 2D_space_time logger that records the information necessary for animation

        Adds a 2D_space_time logger that records the requested number of frames, n_frames.
        The time span of both is from t_min to t_max. The default orientation is in the zy
        plane, but this can be chosen along with the desired resolution.
        Errors can happen with too many empty frames, so it is recommended to set
        t_min close to the first scattering time.
        Select the appropriate animation data from simulation output and use
        plotter.make_animation with a filename to save the animation as a gif.

        Parameters
        ----------

        t_min : float
            Lowest time recorded in [s]

        t_max : float
            Highest time recorded in [s]

        n_frames : int
            Number of frames in space 2D time logger

        d1 : str
            First direction of space 2D time logger, "x", "y" or "z"

        n1 : int
            Number of bins in first direction

        d2 : str
            Second direction of space 2D time logger, "x", "y" or "z"

        n2 : int
            Number of bins in second direction
        """

        if d1 == d2:
            raise RuntimeError("Cant have both d1 and d2 along the same axis.")

        if d1 + d2 in self.animation_loggers:
            raise RuntimeError("Can only add one animation logger with the same axis.")

        self.animation_loggers.append(d1 + d2)

        # find outer dimensions:
        mon_x, mon_y_low, mon_y_high, mon_z = self.find_cryostat_dimensions()

        name = self.name + "_logger_space_" + d1 + d2 + "_time"
        ani_logger = self.instr.add_component(name, "Union_logger_2D_space_time")
        ani_logger.set_AT([0, 0, 0], RELATIVE=self.origin)
        ani_logger.filename = '"' + name + '.dat"'
        ani_logger.time_bins = int(n_frames)
        ani_logger.time_min = t_min
        ani_logger.time_max = t_max

        ani_logger.D_direction_1 = '"' + d1 + '"'
        ani_logger.n1 = int(n1)
        if d1 == "x":
            ani_logger.D1_min = -mon_x
            ani_logger.D1_max = mon_x
        elif d1 == "y":
            ani_logger.D1_min = mon_y_low
            ani_logger.D1_max = mon_y_high
        elif d1 == "z":
            ani_logger.D1_min = -mon_z
            ani_logger.D1_max = mon_z
        else:
            raise RuntimeError("Dimension: '" + d1 + "' not recoignized, must be x, y or z.")

        ani_logger.D_direction_2 = '"' + d2 + '"'
        ani_logger.n2 = int(n2)
        if d2 == "x":
            ani_logger.D2_min = -mon_x
            ani_logger.D2_max = mon_x
        elif d2 == "y":
            ani_logger.D2_min = mon_y_low
            ani_logger.D2_max = mon_y_high
        elif d2 == "z":
            ani_logger.D2_min = -mon_z
            ani_logger.D2_max = mon_z
        else:
            raise RuntimeError("Dimension: '" + d2 + "' not recoignized, must be x, y or z.")

    def add_y_plane(self, value):
        """
        Adds a y plane to database, checking if it is already present

        If duplicates occur, errors would happen in the Union algorithm
        """

        for y_plane in self.used_y_planes:
            if abs(y_plane - value) < 1E-7:
                raise RuntimeError("Two planes overlap with the same y value.")

        self.used_y_planes.append(value)

    def add_radius(self, value):
        """
        Adds a radius to databse, checking if it is already present

        If duplicates occur, errors would happen in the Union algorithm
        """

        for radius in self.used_radius_values:
            if abs(radius - value) < 1E-7:
                raise RuntimeError("The radius of two cylinders are almost equal.")

        self.used_radius_values.append(value)

    def build(self, include_master=True):
        """
        Assigns priorities to the internal components of the cryostat

        The build method must be called after all layers and windows are added
        in order for priorities to be assigned according to the priority window
        given at class initialization. It is optional to include the Union_master
        with the build method, if it is not included it must be manually provided
        later.

        Keyword arguments
        -----------------

        inclde_master : bool
            If True a Union_master component is added to the instrument file
        """

        union_component_list = []

        for layer in self.layers:
            union_component_list += layer.union_components

        priorities = np.linspace(self.max_priority, self.min_priority, len(union_component_list))

        for component, priority in zip(union_component_list, priorities):
            component.priority = priority

        if include_master:
            master = self.instr.add_component(self.name + "_master", "Union_master")




import io
import os
import unittest
import unittest.mock
import matplotlib as plt
import threading

from mcstasscript.interface import instr, functions
from mcstasscript.jb_interface.simulation_interface import SimInterface


class FakeChange:
    def __init__(self, new=None, old=None, name=None):
        self.new = new
        self.old = old
        self.name = name


def setup_complex_instrument():
    """
    Sets up guide system with two guides that are placed next to one
    another with separate entrances but converge at the end.

    It attempts to use as many McStas keywords and features as possible.
    """
    Instr = instr.McStas_instr("integration_test_complex",
                               author="test_suite",
                               origin="integration tests")

    Instr.add_parameter("guide_width", value=0.03)
    Instr.add_parameter("guide_length", value=8.0)

    source = Instr.add_component("source", "Source_simple")
    source.xwidth = 0.1
    source.yheight = 0.01
    source.dist = 1.5
    source.focus_xw = "3*guide_width"
    source.focus_yh = 0.05
    source.E0 = 5.0
    source.dE = 1.0
    source.flux = 1E10

    Instr.add_declare_var("int", "guide_choice")
    Instr.add_declare_var("double", "source_to_guide_end")
    Instr.append_initialize("source_to_guide_end = 1.5 + guide_length;")

    after_guide = Instr.add_component("after_guide", "Arm",
                                      AT=[0, 0, "source_to_guide_end"],
                                      RELATIVE="source")
    after_guide.append_EXTEND("guide_choice = -1;")

    # Add first slit with component methods
    slit1 = Instr.add_component("slit1", "Slit")
    slit1.set_AT(["1.3*guide_width", 0, 1.5], RELATIVE="source")
    slit1.xwidth = "guide_width"
    slit1.yheight = 0.05
    slit1.append_EXTEND("if (SCATTERED) {")
    slit1.append_EXTEND("  guide_choice = 1;")
    slit1.append_EXTEND("}")
    slit1.set_GROUP("entrance_slits")

    # Add second slit with set_parameters
    slit2 = Instr.add_component("slit2", "Slit")
    slit2.set_AT(["-1.3*guide_width", 0, 1.5])
    slit2.set_RELATIVE("source")
    slit2.set_parameters(xwidth="guide_width", yheight=0.05)
    slit2.append_EXTEND("if (SCATTERED) {")
    slit2.append_EXTEND("  guide_choice = 2;")
    slit2.append_EXTEND("}")
    slit2.set_GROUP("entrance_slits")

    select1 = Instr.add_component("select1", "Arm", RELATIVE="after_guide")
    select1.set_JUMP("select2 WHEN guide_choice == 2")

    guide1 = Instr.add_component("guide1", "Guide_gravity")
    guide1.set_AT([0, 0, 0.1], RELATIVE="slit1")
    guide1.set_ROTATED([0, "-RAD2DEG*atan(0.5*guide_width/guide_length)", 0],
                       RELATIVE="slit1")
    guide1.w1 = "guide_width"
    guide1.w2 = "1.3*guide_width"
    guide1.h1 = 0.05
    guide1.h2 = 0.05
    guide1.l = "guide_length"
    guide1.m = 4
    guide1.G = -9.82

    select2 = Instr.add_component("select2", "Arm", RELATIVE="after_guide")
    select2.set_JUMP("done WHEN guide_choice == 1")

    guide2 = Instr.add_component("guide2", "Guide_gravity")
    guide2.set_AT([0, 0, 0.1], RELATIVE="slit2")
    guide2.set_ROTATED([0, "RAD2DEG*atan(0.5*guide_width/guide_length)", 0],
                       RELATIVE="slit2")
    guide2.w1 = "guide_width"
    guide2.w2 = "1.3*guide_width"
    guide2.h1 = 0.05
    guide2.h2 = 0.05
    guide2.l = "guide_length"
    guide2.m = 4
    guide2.G = -9.82

    guide2.set_SPLIT = 2

    Instr.add_component("done", "Arm", RELATIVE="after_guide")

    PSD1 = Instr.add_component("PSD_1D_1", "PSDlin_monitor")
    PSD1.set_AT([0, 0, 0.2], RELATIVE="after_guide")
    PSD1.xwidth = 0.1
    if Instr.mccode_version > 2:
        PSD1.nbins = 100
    else:
        PSD1.nx = 100
    PSD1.yheight = 0.03
    PSD1.filename = "\"PSD1.dat\""
    PSD1.restore_neutron = 1
    PSD1.set_WHEN("guide_choice == 1")

    PSD2 = Instr.add_component("PSD_1D_2", "PSDlin_monitor")
    PSD2.set_AT([0, 0, 0.2], RELATIVE="after_guide")
    PSD2.xwidth = 0.1
    if Instr.mccode_version > 2:
        PSD2.nbins = 100
    else:
        PSD2.nx = 100
    PSD2.yheight = 0.03
    PSD2.filename = "\"PSD2.dat\""
    PSD2.restore_neutron = 1
    PSD2.set_WHEN("guide_choice == 2")

    PSD = Instr.add_component("PSD_1D", "PSDlin_monitor")
    PSD.set_AT([0, 0, 0.2], RELATIVE="after_guide")
    PSD.xwidth = 0.1
    if Instr.mccode_version > 2:
        PSD.nbins = 100
    else:
        PSD.nx = 100
    PSD.yheight = 0.03
    PSD.filename = "\"PSD_all.dat\""
    PSD.restore_neutron = 1

    Instr.append_finally("guide_choice = -1;")

    return Instr


class TestComplexInstrument(unittest.TestCase):
    """
    Integration test of a full instrument with McStas simulation
    performed by the system. The configuration file needs to be set up
    correctly in order for these tests to succeed.
    """
    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_complex_instrument_run(self, mock_stdout):
        """
        Test parameters can be controlled through McStasScript.  Here
        a slit is moved to one side and the result is verified.
        """
        CURRENT_DIR = os.getcwd()
        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        os.chdir(THIS_DIR)

        Instr = setup_complex_instrument()

        data = Instr.run_full_instrument(foldername="integration_test_complex",
                                         ncount=2E6, mpi=2,
                                         increment_folder_name=True,
                                         parameters={"guide_width": 0.03,
                                                     "guide_length": 8.0})

        os.chdir(CURRENT_DIR)

        intensity_data_pos = functions.name_search("PSD_1D_1", data).Intensity
        sum_outside_beam = sum(intensity_data_pos[0:50])
        sum_inside_beam = sum(intensity_data_pos[51:99])
        self.assertTrue(1000*sum_outside_beam < sum_inside_beam)

        intensity_data_neg = functions.name_search("PSD_1D_2", data).Intensity
        sum_outside_beam = sum(intensity_data_neg[51:99])
        sum_inside_beam = sum(intensity_data_neg[0:50])
        self.assertTrue(1000*sum_outside_beam < sum_inside_beam)

        intensity_data_all = functions.name_search("PSD_1D", data).Intensity
        sum_outside_beam = sum(intensity_data_all[49:51])
        sum_inside_beam = (sum(intensity_data_all[0:45])
                           + sum(intensity_data_all[56:99]))
        self.assertTrue(1000*sum_outside_beam < sum_inside_beam)

        # Could have the plot window up for some short time
        # Need to use plt.draw instead of plt.show in plotter
        # plotter.make_sub_plot(data)
        # time.sleep(10)
        # plt.close()

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_complex_instrument_interface(self, mock_stdout):
        """
        Test that a simulation can be performed through the simulation
        interface, or as close as I can through scripting.
        Need to join the simulation thread to the main thread in order
        to wait for the completion as it is performed in a new thread.
        """
        CURRENT_DIR = os.getcwd()
        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        os.chdir(THIS_DIR)

        Instr = setup_complex_instrument()

        interface = SimInterface(Instr)
        interface.show_interface()

        change = FakeChange()

        """
        interface.run_simulation_thread(change)

        for thread in threading.enumerate():
            if thread.name != "MainThread":
                thread.join()
        """

        interface.run_simulation_live(change)

        data = interface.plot_interface.data

        os.chdir(CURRENT_DIR)

        print(data)

        intensity_data_pos = functions.name_search("PSD_1D_1", data).Intensity
        sum_outside_beam = sum(intensity_data_pos[0:50])
        sum_inside_beam = sum(intensity_data_pos[51:99])
        self.assertTrue(1000 * sum_outside_beam < sum_inside_beam)

        intensity_data_neg = functions.name_search("PSD_1D_2", data).Intensity
        sum_outside_beam = sum(intensity_data_neg[51:99])
        sum_inside_beam = sum(intensity_data_neg[0:50])
        self.assertTrue(1000 * sum_outside_beam < sum_inside_beam)

        intensity_data_all = functions.name_search("PSD_1D", data).Intensity
        sum_outside_beam = sum(intensity_data_all[49:51])
        sum_inside_beam = (sum(intensity_data_all[0:45])
                           + sum(intensity_data_all[56:99]))
        self.assertTrue(1000 * sum_outside_beam < sum_inside_beam)


if __name__ == '__main__':
    unittest.main()



import os
from setuptools import setup, find_packages
from glob import glob

# Get version number
here = os.path.abspath(os.path.dirname(__file__))
version_path = os.path.join(here, "mcstasscript", "_version.py")
version = {}
with open(version_path) as fp:
    exec(fp.read(), version)
found_version = version['__version__']
print("Version read from file:", found_version)

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
     name='McStasScript',
     version=found_version,
     author="Mads Bertelsen",
     author_email="Mads.Bertelsen@ess.eu",
     description="A python scripting interface for McStas",
     include_package_data=True,
     long_description=long_description,
     long_description_content_type="text/markdown",
     url="https://github.com/PaNOSC-ViNYL/McStasScript",
     install_requires=['numpy', 'matplotlib', 'PyYAML', 'ipywidgets', 'libpyvinyl'],
     packages=find_packages(),
     classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: GNU General Public License (GPL)",
         "Operating System :: OS Independent",
         "Topic :: Scientific/Engineering"
     ],
 )



from libpyvinyl.Parameters.Parameter import Parameter
from mcstasscript.helper.mcstas_objects import DeclareVariable

def component_description(component):
    """
    Returns string of information about the component

    Includes information on required parameters if they are not yet
    specified. Information on the components are added when the
    class is used as a superclass for classes describing each
    McStas component. Uses mathtext for bold and italics.
    """
    string = ""

    if len(component.c_code_before) > 1:
        string += component.c_code_before + "\n"
    if len(component.comment) > 1:
        string += "// " + component.comment + "\n"
    if component.SPLIT != 0:
        string += "SPLIT " + str(component.SPLIT) + " "
    string += "COMPONENT " + str(component.name)
    string += " = $\\bf{" + str(component.component_name).replace("_", "\\_") + "}$\n"
    for key in component.parameter_names:
        val = getattr(component, key)
        parameter_name = key
        if val is not None:
            unit = ""
            if key in component.parameter_units:
                unit = "[" + component.parameter_units[key] + "]"
            if isinstance(val, Parameter):
                val_string = val.name
            elif isinstance(val, DeclareVariable):
                val_string = val.name
            else:
                val_string = str(val)

            value = "$\\bf{" + val_string.replace("_", "\\_").replace('\"', "''").replace('"', "\''") + "}$"
            string += "  $\\bf{" + parameter_name.replace("_", "\\_") + "}$"
            string += " = " + value + " " + unit + "\n"
        else:
            if component.parameter_defaults[key] is None:
                string += "  $\\bf{" + parameter_name.replace("_", "\\_") + "}$"
                string += " : $\\bf{Required\\ parameter\\ not\\ yet\\ specified}$\n"

    if not component.WHEN == "":
        string += component.WHEN + "\n"

    string += "AT " + str(component.AT_data)
    if component.AT_reference is None:
        string += " $\\it{ABSOLUTE}$\n"
    else:
        string += " RELATIVE $\\it{" + component.AT_reference.replace("_", "\\_") + "}$\n"

    if component.ROTATED_specified:
        string += "ROTATED " + str(component.ROTATED_data)
        if component.ROTATED_reference is None:
            string += " $\\it{ABSOLUTE}$\n"
        else:
            string += " $\\it{" + component.ROTATED_reference.replace("_", "\\_") + "}$\n"

    if not component.GROUP == "":
        string += "GROUP " + component.GROUP + "\n"
    if not component.EXTEND == "":
        string += "EXTEND %{" + "\n"
        string += component.EXTEND + "%}" + "\n"
    if not component.JUMP == "":
        string += "JUMP " + component.JUMP + "\n"
    if len(component.c_code_after) > 1:
        string += component.c_code_after + "\n"

    return string.strip()



import os
import unittest

from mcstasscript.interface.functions import Configurator


def setup_expected_file(test_name):
    THIS_DIR = os.path.dirname(os.path.abspath(__file__))
    expected_file = os.path.join(THIS_DIR, "..", test_name + ".yaml")

    if os.path.isfile(expected_file):
        os.remove(expected_file)
    return expected_file


def setup_configurator(test_name):

    setup_expected_file(test_name)

    return Configurator(test_name)


class TestConfigurator(unittest.TestCase):
    """
    Tests for configurator class that handles yaml configuration file
    """

    def test_simple_initialize(self):
        """
        Tests that initialization happens, new configuration file should be
        written.
        """

        test_name = "test_configuration"
        expected_file = setup_expected_file(test_name)

        # check the file did not exist before testing
        self.assertFalse(os.path.isfile(expected_file))

        # initialize the configurator
        Configurator(test_name)

        # check a new configuration file was made
        self.assertTrue(os.path.isfile(expected_file))

        # remove the testing configuration file
        if os.path.isfile(expected_file):
            os.remove(expected_file)

    def test_default_config(self):
        """
        This tests confirms the content of the default configuration file
        """

        test_name = "test_configuration"
        expected_file = setup_expected_file(test_name)

        # check the file did not exist before testing
        self.assertFalse(os.path.isfile(expected_file))

        my_configurator = Configurator(test_name)

        default_config = my_configurator._read_yaml()

        run = "/Applications/McStas-2.5.app/Contents/Resources/mcstas/2.5/bin/"
        mcstas = "/Applications/McStas-2.5.app/Contents/Resources/mcstas/2.5/"
        mxrun = "/Applications/McXtrace-1.5.app" \
                + "/Contents/Resources/mcxtrace/1.5/mxrun"
        mcxtrace = "/Applications/McXtrace-1.5.app" \
                   + "/Contents/Resources/mcxtrace/1.5/"

        self.assertEqual(default_config["paths"]["mcrun_path"], run)
        self.assertEqual(default_config["paths"]["mcstas_path"], mcstas)
        self.assertEqual(default_config["paths"]["mxrun_path"], mxrun)
        self.assertEqual(default_config["paths"]["mcxtrace_path"], mcxtrace)
        self.assertEqual(default_config["other"]["characters_per_line"], 85)

        # remove the testing configuration file
        if os.path.isfile(expected_file):
            os.remove(expected_file)

    def test_yaml_write(self):
        """
        This test checks that writing to the configuration file works
        """
        test_name = "test_configuration"
        my_configurator = setup_configurator(test_name)

        config = my_configurator._read_yaml()

        config["new_field"] = 123
        config["paths"]["new_path"] = "/test/path/"

        my_configurator._write_yaml(config)

        new_config = my_configurator._read_yaml()

        self.assertEqual(new_config["other"]["characters_per_line"], 85)
        self.assertEqual(new_config["new_field"], 123)
        self.assertEqual(new_config["paths"]["new_path"], "/test/path/")

        # remove the testing configuration file
        setup_expected_file(test_name)

    def test_set_mcrun_path(self):
        """
        This test checks that setting the mcrun path works
        """
        test_name = "test_configuration"
        my_configurator = setup_configurator(test_name)

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        dummy_mcstas_path = os.path.join(THIS_DIR, "dummy_mcstas")

        my_configurator.set_mcrun_path(dummy_mcstas_path)

        new_config = my_configurator._read_yaml()

        self.assertEqual(new_config["paths"]["mcrun_path"], dummy_mcstas_path)

        # remove the testing configuration file
        setup_expected_file(test_name)

    def test_set_mcstas_path(self):
        """
        This test checks that setting the mcstas path works
        """
        test_name = "test_configuration"
        my_configurator = setup_configurator(test_name)

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        dummy_mcstas_path = os.path.join(THIS_DIR, "dummy_mcstas")

        my_configurator.set_mcstas_path(dummy_mcstas_path)

        new_config = my_configurator._read_yaml()

        self.assertEqual(new_config["paths"]["mcstas_path"],
                         dummy_mcstas_path)

        # remove the testing configuration file
        setup_expected_file(test_name)

    def test_set_mxrun_path(self):
        """
        This test checks that setting the mxrun path works
        """
        test_name = "test_configuration"
        my_configurator = setup_configurator(test_name)

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        dummy_mcstas_path = os.path.join(THIS_DIR, "dummy_mcstas")

        my_configurator.set_mxrun_path(dummy_mcstas_path)

        new_config = my_configurator._read_yaml()

        self.assertEqual(new_config["paths"]["mxrun_path"], dummy_mcstas_path)

        # remove the testing configuration file
        setup_expected_file(test_name)

    def test_set_mcxtrace_path(self):
        """
        This test checks that setting the mcxtrace path works
        """
        test_name = "test_configuration"
        my_configurator = setup_configurator(test_name)

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        dummy_mcstas_path = os.path.join(THIS_DIR, "dummy_mcstas")

        my_configurator.set_mcxtrace_path(dummy_mcstas_path)

        new_config = my_configurator._read_yaml()

        self.assertEqual(new_config["paths"]["mcxtrace_path"],
                         dummy_mcstas_path)

        # remove the testing configuration file
        setup_expected_file(test_name)

    def test_set_line_length(self):
        """
        This test checks that setting the line length works
        """
        test_name = "test_configuration"
        my_configurator = setup_configurator(test_name)

        my_configurator.set_line_length(123)

        new_config = my_configurator._read_yaml()

        self.assertEqual(new_config["other"]["characters_per_line"], 123)

        # remove the testing configuration file
        setup_expected_file(test_name)


if __name__ == '__main__':
    unittest.main()



from mcstasscript.instr_reader.util import SectionReader


class TraceReader(SectionReader):
    """
    Reads the trace section of a McStas instrument file. For each
    component a McStasScript component instance is created and the
    parameters/keywords are applied to this instance. When the next
    component is found, the previous component is written to the
    python file for reproduction of the McStas instrument.
    """

    def __init__(self, Instr, write_file, product_filename,
                 get_next_line, return_line):

        super().__init__(Instr, write_file, product_filename,
                         get_next_line, return_line)

        self.current_component = None
        self.in_component_mode = False
        self.EXTEND_mode = False
        self.component_copy_target = None
        self.SPLIT = 0
        self.stored_include = None

    def sanitize_line(self, line):
        """
        Removes comments, the starting blok and newline characters
        """

        line = line.strip()

        # Remove comments
        if "//" in line:
            line = line.split("//", 1)[0].strip()

        if line.startswith("TRACE"):
            line = line.split("TRACE", 1)[1].strip()

        if "/*" in line:
            if "*/" in line:
                line = line.split("/*", 1)[0] + line.split("*/", 1)[1]
            else:
                line = line.split("/*", 1)[0]

        # Remove newline at the end of the line
        if line.endswith("\n"):
            line = line[:-1]

        return line

    def read_trace_line(self, line):
        """
        Reads line of McStas file from TRACE section. Has the responsibility
        of setting continue_trace to false when finding the end of the TRACE
        section. May take extra lines through get_new_line if statements are
        spaced out over several lines.
        """

        continue_trace = True

        # Find stop characeters
        if line.startswith("FINALLY"):
            continue_trace = False

        if line.startswith("END"):
            continue_trace = False

        if line.strip().startswith("%include") or line.strip().startswith("#include"):
            # Handle include statement and attatch it to a component
            if self.current_component is not None:
                c_code_after = self.current_component.c_code_after + line + "\n"
                self.current_component.set_c_code_after(c_code_after)
            else:
                # If the include statement is before the first component,
                # it is saved and attatched to the next component
                line = line.replace('"', "\\\"")
                self.stored_include = line.strip()

        # If the line is just a new line quit
        if line == "\n" or line == "":
            return continue_trace

        line = self.sanitize_line(line)

        # Handle keywords that appear before components
        if line.startswith("SPLIT"):
            # Read split and save for the next component
            line = line.split("SPLIT", 1)[1].strip()
            if line.startswith("COMPONENT"):
                # Default split without indicating amount
                self.SPLIT = ""
            else:
                try:
                    self.SPLIT = int(line.split(" ", 1)[0].strip())
                except:
                    #self.SPLIT = "\"" + line.split(" ", 1)[0].strip() + "\""
                    self.SPLIT = line.split(" ", 1)[0].strip()

                if " " in line:
                    # If the line continues, remove the SPLIT number
                    line = line.split(" ", 1)[1].strip()

        # Read component definition (can be over several lines)
        if line.startswith("COMPONENT") or line.startswith("REMOVABLE COMPONENT"):
            # Start ned component, but write the previous component to file first
            if self.stored_include is not None and self.current_component is not None:
                # In case an include statement was stored, include that statement
                self.current_component.set_c_code_before(self.stored_include)
                self.stored_include = None
            # write previous component
            self._write_component_to_py()

            # start new component
            self.in_component_mode = True
            # Assume this is not a copy
            self.component_copy_target = None

            # Remove COMPONENT from line
            if line.startswith("COMPONENT"):
                line = line[9:].strip()
            elif line.startswith("REMOVABLE COMPONENT"):
                line = line[19:].strip()

            # Add new lines until the entire component definition is found
            full_component_line = False
            while not full_component_line:

                expected_end_parenthesis = 1 + line.count("COPY")

                if line.count("(") >= expected_end_parenthesis:
                    full_component_line = True

                if not full_component_line:
                    new_line = self.get_next_line()
                    new_line = self.sanitize_line(new_line)
                    if new_line.startswith("AT") or new_line.startswith("WHEN"):
                        full_component_line = True
                        self.return_line()
                    else:
                        line += new_line

            # Retrieve information from component definition
            instance_name = line.split("=", 1)[0].strip()
            component_name = line.split("=", 1)[1].split("(", 1)[0].strip()
            line = line.split("=", 1)[1].split("(", 1)[1].strip()
            if component_name == "COPY":
                # Copy instance
                self.component_copy_target = line.split(")", 1)[0].strip()

                if line.strip().startswith("("):
                    line = line.split("(", 1)[1].strip()

                if self.component_copy_target == "PREVIOUS":
                    # Get the previous component name
                    last_component = self.Instr.get_last_component()
                    self.component_copy_target = last_component.name

                if "(" in instance_name:
                    # Using the copy notation, replace this with meaningful replacement
                    base_name = self.component_copy_target + "_copy"
                    name = base_name
                    comp_names = [x.name for x in self.Instr.component_list]
                    index = 0
                    while name in comp_names:
                        name = base_name + "_" + str(index)
                        index += 1

                    instance_name = name

                self.current_component = self.Instr.copy_component(instance_name,
                                                                   self.component_copy_target)

            else:
                # Normal component instance
                self.current_component = self.Instr.add_component(instance_name,
                                                                  component_name)

            # In case there are no parameters, stop in_component_mode
            if line.startswith(")"):
                self.in_component_mode = False
                line = line.split(")", 1)[1]

            if self.SPLIT != 0:
                self.current_component.set_SPLIT(self.SPLIT)
                self.SPLIT = 0

        # In case of COPY, there can be empty parameter lists
        if self.in_component_mode:  # and self.component_copy_target != None:
            # Check if this line starts WHEN or AT
            if line.strip().startswith("AT"):
                self.in_component_mode = False
            if line.strip().startswith("WHEN"):
                self.in_component_mode = False
            # If none of these occur, read the parameters

        # In component mode reads parameters of each new line read
        if self.in_component_mode:

            par_line = line
            # check for parameters
            if par_line.strip().startswith("("):
                par_line = line.split("(", 1)[1]

            # _in_func like python in, but does not look inside parenthesis
            if self._in_func(line, ")"):
                self.in_component_mode = False
                par_line = self._split_func(line, ")", 1)[0]
                line = self._split_func(line, ")", 1)[1]

            # All parameters found saved in dictionary
            par_dict = {}

            """
            A parameter line can contain a comma for separating parameters or
            inside of a string. This piece of code finds the next comma and
            checks if there is an equal number of quotation marks in the first
            part, if not it increases the read part of the line to the next
            comma. In this way commas in strings do not separate parameters
            in the component input.
            """
            while len(par_line) > 0:
                # find the next parameter expression
                if self._in_func_brack(par_line, ","):
                    # start of expression to evaluate
                    par_exp = par_line.split(",", 1)[0].strip()
                    par_exp = self._split_func_brack(par_line, ",", 1)[0]
                    # remove the part already taken from the par_line
                    par_line = self._split_func_brack(par_line, ",", 1)[1]
                    # The length of quotation_split will be one more than
                    #  the number of quotation marks in par_exp
                    quotation_split = par_exp.split('"')
                    while (len(quotation_split) - 1) % 2 != 0:
                        # There is an uneven number of quotation marks
                        par_exp += ","
                        if "," in par_line:
                            # include up to the next comma in par_exp
                            par_exp += par_line.split(",", 1)[0]
                            # remove the part of the par_line added to par_exp
                            par_line = par_line.split(",", 1)[1]
                        else:
                            # no commas left, must be end of par_line
                            par_exp += par_line.strip()
                            par_line = ""

                        quotation_split = par_exp.split('"')
                else:
                    # last parameter
                    par_exp = par_line
                    par_line = ""

                if "=" in par_exp:
                    par_name = par_exp.split("=", 1)[0].strip()
                    par_value = par_exp.split("=", 1)[1].strip()

                    par_dict[par_name] = par_value

            # Set all found parameters in the component
            self.current_component.set_parameters(par_dict)

        # Read keywords given after parameters but before position (WHEN)
        if line.strip().upper().startswith("WHEN"):
            if "(" in line:
                line = line.split("(", 1)[1].strip()
                # need to find the closing parenthesis
                parenthesis_counter = 1
                character_index = -1
                for character in line:
                    character_index += 1
                    if character == "(":
                        parenthesis_counter += 1
                    if character == ")":
                        parenthesis_counter -= 1
                    if parenthesis_counter == 0:
                        end_index = character_index
                        break

                WHEN_statement = line[:character_index]
                WHEN_statement = WHEN_statement.replace('"', "\\\"")
                line = line[character_index+1:].strip()
            else:
                # WHEN statement that does not use parenthesis
                if "AT" in line:
                    WHEN_statement = line.split("AT", 1)[0]
                    line = "AT" + line.split("AT", 1)[1]
                else:
                    WHEN_statement = line.split("WHEN ", 1)[1]
                    line = ""

            self.current_component.set_WHEN(WHEN_statement)

        # Read component position
        if line.strip().startswith("AT"):
            # read AT statement
            line = line.split("(", 1)[1].strip()
            AT_data = []
            AT_data.append(self._split_func(line, ",", 1)[0])
            line = self._split_func(line, ",", 1)[1]
            AT_data.append(self._split_func(line, ",", 1)[0])
            line = self._split_func(line, ",", 1)[1]
            AT_data.append(self._split_func(line, ")", 1)[0])
            line = self._split_func(line, ")", 1)[1]

            if line.strip().startswith("ABSOLUTE"):
                line = line.split(" ", 1)[1].strip()
                relative_name = "ABSOLUTE"
                # The line can continue, remove the used part
                if " " in line.strip():
                    line = line.strip().split(" ", 1)[1].strip()
                else:
                    line = ""
            elif line.strip().startswith("RELATIVE"):
                line = line.strip().split(" ", 1)[1].strip()
                if " " in line:
                    relative_name = line.split(" ", 1)[0].strip()
                else:
                    relative_name = line.strip()
                # The line can continue, remove the used part
                if " " in line:
                    line = line.split(" ", 1)[1].strip()
                else:
                    line = ""
            else:
                raise ValueError("Could not read: " + line)

            self.current_component.set_AT(AT_data, RELATIVE=relative_name)

        # Read component rotation
        if line.strip().startswith("ROTATED"):
            # read ROTATED statement
            line = line.split("(", 1)[1].strip()
            ROTATED_data = []

            ROTATED_data.append(self._split_func(line, ",", 1)[0])
            line = self._split_func(line, ",", 1)[1]
            ROTATED_data.append(self._split_func(line, ",", 1)[0])
            line = self._split_func(line, ",", 1)[1]
            ROTATED_data.append(self._split_func(line, ")", 1)[0])
            line = self._split_func(line, ")", 1)[1]

            if line.strip().startswith("ABSOLUTE"):
                relative_name = "ABSOLUTE"
                if " " in line.strip():
                    line = line.strip().split(" ", 1)[1].strip()
                else:
                    line = ""
            elif line.strip().startswith("RELATIVE"):
                line = line.strip().split(" ", 1)[1].strip()
                relative_name = line.split(" ", 1)[0].strip()
                # The line can continue, remove the used part
                if " " in line:
                    line = line.split(" ", 1)[1].strip()
                else:
                    line = ""
            else:
                raise ValueError("Could not read: " + line)

            self.current_component.set_ROTATED(ROTATED_data,
                                               RELATIVE=relative_name)

        # Read keywords after component position (GROUP, EXTEND, JUMP)
        if line.strip().upper().startswith("GROUP"):
            line = line.strip()

            group_name = line.split(" ", 1)[1].strip()
            #group_name = "\"" + group_name + "\""
            group_name = group_name

            line = ""

            self.current_component.set_GROUP(group_name)

        if line.strip().upper().startswith("EXTEND"):
            line = line.split("EXTEND", 1)[1].strip()
            self.EXTEND_mode = True

        if self.EXTEND_mode:
            if "%{" in line:
                line = line.strip().split("%{", 1)[1].strip()

            if "%}" in line:
                line = line.strip().split("%}", 1)[0].strip()
                self.EXTEND_mode = False

            if len(line) > 0 and line != "\n":
                line = line.replace('\\n', "\\\\n")
                line = line.replace('"', "\\\"")
                self.current_component.append_EXTEND(line)

        if line.strip().upper().startswith("JUMP "):
            line = line.strip().split(" ", 1)[1]
            self.current_component.set_JUMP(line)

        if not continue_trace:
            # write last component
            self._write_component_to_py()

        return continue_trace

    def _write_component_to_py(self):
        # code for writing McStasScript python file
        if self.current_component is not None:

            # Write the add_component statement
            write_string = ["\n"]
            write_string.append(self.current_component.name)
            write_string.append(" = ")
            write_string.append(self.instr_name)
            if self.component_copy_target is None:
                write_string.append(".add_component(")
                write_string.append("\"" + self.current_component.name + "\"")
                write_string.append(", ")
                component_name = str(self.current_component.component_name)
                write_string.append("\"" + component_name + "\"")
            else:
                write_string.append(".copy_component(")
                write_string.append("\"" + self.current_component.name + "\"")
                write_string.append(", ")
                write_string.append("\"" + self.component_copy_target + "\"")
            write_string.append(")\n")

            self._write_to_file(write_string)

            # Write all parameters as attribute updates
            for key in self.current_component.parameter_names:
                val = getattr(self.current_component, key)
                if val is not None:
                    write_string = []
                    write_string.append(self.current_component.name)
                    write_string.append(".")
                    write_string.append(key)
                    write_string.append(" = ")

                    try:
                        # No problems if the value is a number
                        float(val)
                    except:
                        # If the value is a string, it needs quotes
                        if '"' in val:
                            # If it already has quotes, these need escapes
                            val = val.replace('"', '\\\"')
                        val = '"' + val + '"'

                    write_string.append(val)
                    write_string.append("\n")

                    self._write_to_file(write_string)

            # Write EXTEND block if present
            if self.current_component.EXTEND != "":
                EXTEND = self.current_component.EXTEND
                EXTEND_lines = EXTEND.split("\n")
                EXTEND_lines = EXTEND_lines[:-1]

                for EXTEND_line in EXTEND_lines:
                    write_string = []
                    write_string.append(self.current_component.name)
                    write_string.append(".append_EXTEND(")
                    write_string.append("\"" + EXTEND_line + "\"")
                    write_string.append(")\n")

                    self._write_to_file(write_string)

            # Write WHEN statement if present
            if self.current_component.WHEN != "":
                write_string = []
                write_string.append(self.current_component.name)
                write_string.append(".set_WHEN(")
                WHEN = self.current_component.WHEN
                WHEN = WHEN.split("(", 1)[1].strip()
                WHEN = WHEN[:-1]
                write_string.append("\"" + WHEN + "\"")
                write_string.append(")\n")

                self._write_to_file(write_string)

            # Write SPLIT if present
            if self.current_component.SPLIT != 0:
                write_string = []
                write_string.append(self.current_component.name)
                write_string.append(".set_SPLIT(")
                if self.current_component.SPLIT == "":
                    write_string.append('""')
                else:
                    if isinstance(self.current_component.SPLIT, str) and self.current_component.SPLIT != "":
                        write_string.append('"' + str(self.current_component.SPLIT) + '"')
                    else:
                        write_string.append(str(self.current_component.SPLIT))
                write_string.append(")\n")

                self._write_to_file(write_string)

            # Write GROUP if present
            if self.current_component.GROUP != "":
                write_string = []
                write_string.append(self.current_component.name)
                write_string.append(".set_GROUP(\"")
                write_string.append(str(self.current_component.GROUP))
                write_string.append("\")\n")

                self._write_to_file(write_string)

            # Write JUMP if present
            if self.current_component.JUMP != "":
                write_string = []
                write_string.append(self.current_component.name)
                write_string.append(".set_JUMP(")
                jump_string = str(self.current_component.JUMP)
                write_string.append("\"" + jump_string + "\"")
                write_string.append(")\n")

                self._write_to_file(write_string)

            # Write c_code_before if present
            if self.current_component.c_code_before != "":
                write_string = []
                write_string.append(self.current_component.name)
                write_string.append(".set_c_code_before(")
                c_code_string = str(self.current_component.c_code_before)
                write_string.append("\"" + c_code_string + "\"")
                write_string.append(")\n")

                self._write_to_file(write_string)

            # Write c_code_after if present
            if self.current_component.c_code_after != "":
                write_string = []
                write_string.append(self.current_component.name)
                write_string.append(".set_c_code_after(")
                c_code_string = str(self.current_component.c_code_after)
                write_string.append("\"" + c_code_string + "\"")
                write_string.append(")\n")

                self._write_to_file(write_string)

            # Write AT
            write_string = []
            write_string.append(self.current_component.name)
            write_string.append(".set_AT(")
            write_string.append(str(self.current_component.AT_data))
            write_string.append(", RELATIVE=")
            if self.current_component.AT_relative == "ABSOLUTE":
                write_string.append("\"" + "ABSOLUTE" + "\"")
            else:
                relative = self.current_component.AT_relative.split(" ")[1]
                write_string.append("\"" + relative + "\"")
            write_string.append(")\n")

            self._write_to_file(write_string)

            # Write ROTATED
            write_string = []
            write_string.append(self.current_component.name)
            write_string.append(".set_ROTATED(")
            write_string.append(str(self.current_component.ROTATED_data))
            write_string.append(", RELATIVE=")
            if self.current_component.ROTATED_relative == "ABSOLUTE":
                write_string.append("\"" + "ABSOLUTE" + "\"")
            else:
                relative = self.current_component.ROTATED_relative.split(" ")
                relative = relative[1]
                write_string.append("\"" + relative + "\"")
            write_string.append(")\n")

            if self.current_component.ROTATED_specified:
                self._write_to_file(write_string)



import copy

from mcstasscript.instrument_diagram.box import ComponentBox
from mcstasscript.instrument_diagram.generate_AT import generate_AT_arrows
from mcstasscript.instrument_diagram.generate_ROTATED import generate_ROTATED_arrows
from mcstasscript.instrument_diagram.generate_JUMP import generate_JUMP_arrows
from mcstasscript.instrument_diagram.generate_target_index import generate_target_index_arrows
from mcstasscript.instrument_diagram.generate_GROUP import generate_GROUP_arrows
from mcstasscript.instrument_diagram.generate_Union import generate_Union_arrows
from mcstasscript.instrument_diagram.canvas import DiagramCanvas
from mcstasscript.instrument_diagnostics.intensity_diagnostics import IntensityDiagnostics


def instrument_diagram(instrument, analysis=False, variable=None, limits=None):
    """
    Plots diagram of components in instrument with RELATIVE connections

    All components in the instrument are shown as text fields and arrows are
    drawn showing the AT RELATIVE and ROTATED RELATIVE connections between
    components. When more advanced features are used, these are displayed
    with arrows on the right side of the diagram, currently JUMP, GROUP and
    use of Union components are visualized.

    Parameters
    ----------

    instrument : McCode_instr
        Instrument object from which the component list is taken
    """

    # Grab components from instrument file and make text box objects
    components = instrument.make_component_subset()

    # Prepare legend
    component_reader = instrument.component_reader
    component_categories = copy.deepcopy(component_reader.component_category)

    absolute_box = ComponentBox("ABSOLUTE")
    component_boxes = [absolute_box]
    component_box_dict = {"ABSOLUTE": absolute_box}
    for component in components:
        box = ComponentBox(component)
        component_boxes.append(box)
        component_box_dict[component.name] = box

    box_names = [x.name for x in component_boxes]

    color_choices = {"AT": "blue", "ROTATED": "red", "JUMP": "black", "GROUP": [0.4, 0.4, 0.4], "Union": "green"}

    # Arrows for the left side of the diagram
    AT_arrows = generate_AT_arrows(components, component_box_dict, box_names, color=color_choices["AT"])
    ROTATED_arrows = generate_ROTATED_arrows(components, component_box_dict, box_names, color=color_choices["ROTATED"])

    # Arrow for the right side of the diagram
    try:
        JUMP_arrows = generate_JUMP_arrows(components, component_box_dict, box_names, color=color_choices["JUMP"])
    except:
        JUMP_arrows = []
        print("Generation of JUMP arrows failed, please send bug report.")

    try:
        target_index_arrows = generate_target_index_arrows(components, component_box_dict,
                                                           box_names, color=color_choices["JUMP"])
    except:
        target_index_arrows = []
        print("Generation of target_index arrows failed, please send bug report.")

    try:
        GROUP_arrows = generate_GROUP_arrows(components, component_box_dict, box_names, color=color_choices["GROUP"])
    except:
        GROUP_arrows = []
        print("Generation of GROUP arrows failed, please send bug report.")

    try:
        Union_arrows = generate_Union_arrows(components, component_box_dict, box_names,
                                             component_categories, color=color_choices["Union"])
    except:
        Union_arrows = []
        print("Generation of Union arrows failed, please send bug report.")

    intensity_diagnostics=None
    if analysis or variable is not None:
        intensity_diagnostics = IntensityDiagnostics(instrument)
        intensity_diagnostics.settings(suppress_output=True)

    # Create canvas
    canvas = DiagramCanvas(AT_arrows + ROTATED_arrows, component_boxes,
                           JUMP_arrows + target_index_arrows + GROUP_arrows + Union_arrows,
                           component_categories=component_categories, colors=color_choices,
                           intensity_diagnostics=intensity_diagnostics, variable=variable,
                           limits=limits)

    # Plot diagram
    canvas.plot()


def find_python_variable_name(error_text, n_levels=1):
    """
    Finds call in stack n_levels above and gets variable name

    This function is used to avoid having to manually specify a name
    in the method or function inputs. For example with a parameters:

    wavelength = instrument.add_parameter("wavelength", value=3)

    can be equivalent to:

    wavelength = instrument.add_parameter(value=3)

    When this function is used to deduct the appropriate name.
    The n_levels is the number of function calls between the user
    call and this function, for each level of nesting add another
    level. An error_text should be given for cases where it was not
    possible to extract the variable name.
    """

    import inspect

    stack = inspect.stack()
    calling_string = ' '.join(stack[n_levels][4])
    if '=' not in calling_string:
        raise NameError(error_text)
    return calling_string.split('=')[0].strip()